#!/usr/bin/env python

"""Information-based generalized Robinson-Foulds distances

References
----------
- https://ms609.github.io/TreeDist/reference/TreeDistance.html
"""

from typing import Tuple, Sequence, Set

# FIXME: support backup method old Python does not support
# from functools import cache
import itertools

from loguru import logger
from scipy.special import factorial2
from scipy.optimize import linear_sum_assignment
import numpy as np
import pandas as pd
from toytree.utils import ToytreeError
from toytree import ToyTree
from toytree.core.apis import TreeDistanceAPI, add_subpackage_method

logger = logger.bind(name="toytree")

####################################################
# Size of Tree Space, Tree Combinatorics, used to calc shared phylo info
#
# >>> _get_n_unrooted_trees(5)  # 105
####################################################


def _get_n_unrooted_trees(size: int) -> int:
    """Return the number of possible unrooted trees for ntips=size."""
    return int(factorial2(2 * size - 5))


def _get_n_rooted_trees(size: int) -> int:
    """Return the number of possible rooted trees for ntips=size."""
    return int(factorial2(2 * size - 3))


def _get_n_trees_matching_split(size_a: int, size_b: int) -> int:
    """Return number of trees that could contain split of size (A, B)"""
    if size_a == 0:
        return _get_n_unrooted_trees(size_b)
    if size_b == 0:
        return _get_n_unrooted_trees(size_a)
    return factorial2(2 * size_a - 3) * factorial2(2 * size_b - 3)


def _get_n_trees_matching_two_splits(ntips: int, size_a1: int, size_a2: int) -> int:
    """Return N trees that could contain A1 and A2 where A1 is subset of A2.

    Based on 'TreesConsistentWithTwoSplits' from TreeDist.
    """
    small = min(size_a1, size_a2)
    large = max(size_a1, size_a2)

    # simple scenarios
    if small == 0:
        return _get_n_trees_matching_split(large, ntips - large)
    if large == ntips:
        return _get_n_trees_matching_split(small, ntips - small)

    # calculate matching
    n_overlap = large - small
    part1 = _get_n_rooted_trees(n_overlap + 1)
    part2 = _get_n_rooted_trees(small)
    part3 = _get_n_rooted_trees(ntips - large)
    return int(part1 * part2 * part3)


#################################################################
# Get matching split info for bipartitions
##################################################################

def _get_two_splits_matching_split_dist(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> float:
    """Return the number of elements that must be moved from one
    subset to another in order to make the two splits identical.
    """
    a1, b1 = (set(i) for i in split1)
    a2, b2 = (set(i) for i in split2)
    ntips = len(a1) + len(b1)
    return ntips - max(len(a1 & a2) + len(b1 & b2), len(a1 & b2) + len(b1 & a2))


#################################################################
# Get Nye similarity metric
##################################################################

def _get_two_splits_nye_similarity(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> float:
    """Return the Nye similarity metric for two splits.

    Considers the elements held in common between subsets of each
    split, and considers both possible alignments of subsets, returning
    only the alignment score of the best.
    """
    a1, b1 = (set(i) for i in split1)
    a2, b2 = (set(i) for i in split2)

    # get alignment score from subset scores and return max alignment
    subset_score_1 = _get_nye_score(a1, a2)
    subset_score_2 = _get_nye_score(b1, b2)
    ali_score_1 = min(subset_score_1, subset_score_2)
    subset_score_1 = _get_nye_score(a1, b2)
    subset_score_2 = _get_nye_score(a2, b1)
    ali_score_2 = min(subset_score_1, subset_score_2)
    return max(ali_score_1, ali_score_2)


def _get_nye_score(set1: Set, set2: Set) -> float:
    """Return the Nye score for two sets"""
    return len(set1 & set2) / len(set1 | set2)


def get_tree_splitwise_nye_similarity(tree: ToyTree) -> float:
    """Return the maximum phylogenetic information in a tree."""
    return sum(1 for i in tree.iter_bipartitions())


#################################################################
# Get phylogenetic info from bipartitions
#
# >>> _get_phylo_info(4, 5)                         # 6.42...
# >>> _get_split_phylo_info((0,1,2,3),(4,5,6,7,8))  # 6.42...
##################################################################

def _get_phylo_prob(size_a: int, size_b: int) -> float:
    """Return prob that binary tree of size X contains split A|B."""
    size_x = size_a + size_b
    trees_w_split = _get_n_trees_matching_split(size_a, size_b)
    return trees_w_split / _get_n_unrooted_trees(size_x)


# @cache
def _get_phylo_info(size_a: int, size_b: int) -> float:
    """Return information of phylo prob in units of bits."""
    if (size_a < 2) or (size_b < 2):
        return 0
    return -np.log2(_get_phylo_prob(size_a, size_b))


# @cache
def _get_split_phylo_info(split: Tuple[Tuple, Tuple]) -> float:
    """Return the phylogenetic info of a split sensu Martin Smith 2020.

    Phy Info is the prob that a randomly sampled binary tree of
    size X contains the split S. It is used in rf-info and returned
    in units of bits using a log base 2.

    The split AB|CDEF occurs in 15/105 of the possible resolutions
    of a six leaf tree, so the prob of seeing it is P(S1) = 15/105,
    and its information content is h(S1) = -log_2(15/105) = 2.81.
    """
    # uses cache func to get stored result for split sizes previously seen.
    return _get_phylo_info(len(split[0]), len(split[1]))


def get_tree_splitwise_phylo_info(tree: ToyTree) -> float:
    """Return the maximum phylogenetic information in a tree."""
    return sum(_get_split_phylo_info(i) for i in tree.iter_bipartitions())


####################################################
# Get phylogenetic info from PAIRS of bipartitions
#
# >>> set1 = (tuple("AB"), tuple("CDEF"))
# >>> set2 = (tuple("ABC"), tuple("DEF"))
# >>> _get_two_splits_shared_phylo_info(set1, set2)  # 1.222...
####################################################

def _get_subset_superset(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> Tuple[Tuple, Tuple]:
    """Return a bipart half (A1, A2) from each bipart so A1 is a
    subset of A2, else A2 will return empty.
    """
    # select smaller half of split1 and superset from split 2
    subset = set(split1[0])
    superset = [i for i in split2 if set(i).issuperset(subset)]
    superset = superset[0] if superset else ()
    if not superset:
        subset = set(split1[1])
        superset = [i for i in split2 if set(i).issuperset(subset)]
        superset = superset[0] if superset else ()
    return tuple(sorted(subset)), superset


def _get_phylo_prob_two_splits(ntips: int, size_a1: int, size_a2: int) -> float:
    """Return the prob two splits exist in a tree of size X tips"""
    ntrees = _get_n_trees_matching_two_splits(ntips, size_a1, size_a2)
    ntotal = _get_n_unrooted_trees(ntips)
    return ntrees / ntotal


# @cache
def _get_phylo_info_two_splits(ntips: int, size_a1: int, size_a2: int) -> float:
    """Return the phylo information in two splits matching in a tree"""
    return -np.log2(_get_phylo_prob_two_splits(ntips, size_a1, size_a2))


def _get_two_splits_shared_phylo_info(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> float:
    """Return the shared phylo info (spi) for two splits in a tree.

    if conflict:
        SPI = h_shared = 0
    else:
        SPI = h_shared = h(S1) + h(S2) - h(S1, S2)

    Examples
    --------
    >>> split1 = ((0, 1, 2), (3, 4, 5, 6, 7))
    >>> split2 = ((0, 1, 2, 3), (4, 5, 6, 7))
    >>> _get_two_splits_shared_phylo_info(split1, split1)  # 5.044
    >>> _get_two_splits_shared_phylo_info(split2, split2)  #
    """
    subset, superset = _get_subset_superset(split1, split2)
    if not superset:
        return 0
    ntips = len(split1[0]) + len(split1[1])
    hs1 = _get_split_phylo_info(split1)
    hs2 = _get_split_phylo_info(split2)
    # check: returns S1S2 = S1 when S1=S2
    hs12 = _get_phylo_info_two_splits(ntips, len(subset), len(superset))
    return (hs1 + hs2) - hs12


def _get_two_splits_different_phylo_info(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> float:
    """Return the shared phylo info (spi) for two splits in a tree.

    if conflict:
        SPI = h_shared = 0
    else:
        SPI = h_shared = h(S1) + h(S2) - h(S1, S2)
    """
    hs1 = _get_split_phylo_info(split1)
    hs2 = _get_split_phylo_info(split2)
    subset, superset = _get_subset_superset(split1, split2)
    if not superset:
        return hs1 + hs2
    ntips = len(split1[0]) + len(split1[1])
    hs12 = _get_phylo_info_two_splits(ntips, len(subset), len(superset))
    hshared = (hs1 + hs2) - hs12
    return hs12 - hshared


####################################################
# Information theoretic alternative to matching split dist
####################################################

def _get_two_splits_matching_split_phylo_info(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> float:
    """Return the phylogenetic information of matching splits.

    In the matching split distance, m represents a simple count of
    the number of shared taxa. An alternative is to measure the
    phylogenetic information content of the largest split consistent
    with S1 and S2:
    """
    a1, b1 = (set(i) for i in split1)
    a2, b2 = (set(i) for i in split2)
    info1 = _get_phylo_info(len(a1 & a2), len(b1 & b2))
    info2 = _get_phylo_info(len(a1 & b2), len(b1 & a2))
    return max(info1, info2)


####################################################
# Get clustering entropy of bipartitions
####################################################

# @cache
def _entropy(*prob: float) -> float:
    """Return sum of entropies where e = prob * log2(prob)"""
    return -sum(p * np.log2(p) if p else 0 for p in prob)


def _get_split_entropy(split: Tuple[Tuple, Tuple]) -> float:
    """Return the clustering entropy associated with a split S.

    If a split S = A|B, then PCl(A) denotes the probability that a
    randomly selected leaf belongs to A: `PCl(A) = |A| / |X|`.
    Entropy(S) = −PCl(A) * log(PCl(A)) − PCl(B) * log(PCl(B))
    """
    len_a, len_b = [len(i) for i in split]
    ntips = len_a + len_b
    return _entropy(len_a / ntips, len_b / ntips)


def get_tree_splitwise_entropy(tree: ToyTree) -> float:
    """Return the maximum entropy of a tree"""
    return sum(_get_split_entropy(i) for i in tree.iter_bipartitions())


####################################################
# Get clustering entropy from PAIRS of bipartitions
####################################################

def _get_two_splits_entropy(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> Tuple[float, float, float, float, float]:
    """Return entropy info for two splits.

    Returns h_1, h_2, h_joint, h_shared, h_dist
    """
    ntips = len(split1[0]) + len(split1[1])
    sets = [set(i) for i in (split1[0], split1[1], split2[0], split2[1])]
    overlaps = (i & j for (i, j) in itertools.combinations(sets, 2))
    overlaps = [i for i in overlaps if i]

    h_1 = _get_split_entropy(split1)
    h_2 = _get_split_entropy(split2)
    h_joint = _entropy(*[len(i) / ntips for i in overlaps])
    h_shared = h_1 + h_2 - h_joint
    h_dist = h_joint - h_shared
    return h_1, h_2, h_joint, h_shared, h_dist


def _get_two_splits_entropy_info(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> float:
    return _get_two_splits_entropy(split1, split2)[3]


def _get_two_splits_entropy_info_dist(
    split1: Tuple[Tuple, Tuple],
    split2: Tuple[Tuple, Tuple],
) -> float:
    return _get_two_splits_entropy(split1, split2)[4]


####################################################################
# Get matching of bipartitions based on split similarity scores,
# i.e., shared phylo info (spi) or mutual clustering info (msi)
####################################################################

def _get_split_matching(
    biparts1: Sequence[Tuple[Tuple, Tuple]],
    biparts2: Sequence[Tuple[Tuple, Tuple]],
    split_similarity_metric: str = "spi",
) -> float:
    """Return a paired similarity matrix and optimal matching.

    An optimal matching can be found by building a table of pairwise
    similarity scores among all bipartitions and then solving a
    **linear sum assignment problem**. Similarity scores are measured
    by the *shared phylogenetic info* algorithm of Martin Smith, and
    the linear assignment is performed by scipy using the Hungarian
    algorithm.

    Parameters
    ----------
    split_similarity_metric: str
        "spi": shared phylogenetic information.
        "mci": mutual clustering information.
        "msi": matching split phylogenetic information.
        "ms": matching split distance.

    References
    ----------
    - ...
    """
    # split similarity function
    if split_similarity_metric == "mci":
        _get_split_similarity = _get_two_splits_entropy_info
        maximize = True
    elif split_similarity_metric == "spi":
        _get_split_similarity = _get_two_splits_shared_phylo_info
        maximize = True
    elif split_similarity_metric == "ms":
        _get_split_similarity = _get_two_splits_matching_split_dist
        maximize = False
    elif split_similarity_metric == "msi":
        _get_split_similarity = _get_two_splits_matching_split_phylo_info
        maximize = True
    elif split_similarity_metric == "nye":
        _get_split_similarity = _get_two_splits_nye_similarity
        maximize = True
    else:
        raise ToytreeError(
            "split similarity metric must be in "
            "('mci', 'spi', 'ms', 'msi', 'nye')")

    # get partitions in their input order.
    nb1 = len(biparts1)
    nb2 = len(biparts2)

    # get matric of split similarity measures
    arr = np.zeros(shape=(nb1, nb2))
    for idx, bip1 in enumerate(biparts1):
        for jdx, bip2 in enumerate(biparts2):
            arr[idx, jdx] = _get_split_similarity(bip1, bip2)

    # get linear solution pairing splits
    indices = linear_sum_assignment(arr, maximize=maximize)
    return arr, indices


def _report_matching(
    tree1: ToyTree,
    tree2: ToyTree,
    split_similarity_metric: str = "spi",
) -> Tuple["Canvas", "Cartesian", "Mark"]:
    """View the matching of splits between two trees.

    This method is used only for testing/validation, and to learn
    about generalized RF distance methods.

    Returns
    -------
    A toyplot Canvas with 3 Cartesian axes: [Matrix] [Tree1] [Tree2].
    """
    # get the matching
    biparts1 = list(tree1.iter_bipartitions())
    biparts2 = list(tree2.iter_bipartitions())
    pairs, indices = _get_split_matching(biparts1, biparts2, split_similarity_metric)

    nb1 = len(biparts1)
    nb2 = len(biparts2)
    data = pd.DataFrame(
        pairs,
        index=range(tree1.ntips, tree1.ntips + nb1),
        columns=range(tree2.ntips, tree2.ntips + nb2),
    )
    print(f"\n# paired split scores labeled by Node idx\n{data}")

    # add size to align indices with Node labels
    # arr1 += nb1
    # arr2 += nb2
    print("\n# matched split indices, tip labels, and scores")
    for i, j in zip(*indices):
        bip1 = biparts1[i]
        bip1 = f"{','.join(bip1[0])}|{','.join(bip1[1])}"
        bip2 = biparts2[j]
        bip2 = f"{','.join(bip2[0])}|{','.join(bip2[1])}"
        score = pairs[i, j]
        if score:
            print(f"{i}\t{j}\t{bip1} => {bip2}\t{score:.3f}")
        else:
            print(f"{i}\t{j}\t{bip1} .. {bip2}\t{score:.3f}")


####################################################################
# Functions using matched bipartitions
# Reference: https://ms609.github.io/TreeDist/reference/TreeDistance.html
####################################################################

def get_trees_nye_similarity(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return the Nye similarity between two trees.

    References
    ----------
    - Nye TMW, Liò P, Gilks WR (2006). “A novel algorithm and
    web-based tool for comparing two alternative phylogenetic trees.”
    Bioinformatics, 22(1), 117--119. doi: 10.1093/bioinformatics/bti720.
    """
    biparts1 = list(tree1.iter_bipartitions())
    biparts2 = list(tree2.iter_bipartitions())
    arr, indices = _get_split_matching(biparts1, biparts2, "nye")
    nye = arr[indices].sum()
    if normalize:
        ind_info = sum(get_tree_splitwise_nye_similarity(i) for i in (tree1, tree2))
        return nye / (ind_info / 2)
    return nye


def get_trees_nye_dist(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return the Nye similarity distance between two trees.
    """
    nye = get_trees_nye_similarity(tree1, tree2, normalize=False)
    ind_info = sum(get_tree_splitwise_nye_similarity(i) for i in (tree1, tree2))
    if normalize:
        return (ind_info - (2 * nye)) / ind_info
    return ind_info - (2 * nye)


def get_trees_matching_split_dist(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return the matching split distance between two trees.

    Distance is the number of elements that must be moved from one
    subset to another in order to make the two splits identical.
    """
    biparts1 = list(tree1.iter_bipartitions())
    biparts2 = list(tree2.iter_bipartitions())
    arr, indices = _get_split_matching(biparts1, biparts2, "ms")
    msd = arr[indices].sum()
    if normalize:
        logger.warning("no normalization method for matching split distance.")
    return msd


def get_trees_matching_split_info(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return the matching split info distance between two trees.

    Matching uses the number of elements that must be moved from one
    subset to another in order to make the two splits identical, but,
    then the phylo info is returned for each match.
    """
    biparts1 = list(tree1.iter_bipartitions())
    biparts2 = list(tree2.iter_bipartitions())
    arr, indices = _get_split_matching(biparts1, biparts2, "msi")
    msi = arr[indices].sum()
    if normalize:
        ind_info = sum(get_tree_splitwise_phylo_info(i) for i in (tree1, tree2))
        return msi / (ind_info / 2)
    return msi


def get_trees_matching_split_info_dist(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return the matching split info distance between two trees.

    Matching uses the number of elements that must be moved from one
    subset to another in order to make the two splits identical, but,
    then the phylo info is returned for each match.
    """
    msi = get_trees_matching_split_info(tree1, tree2, normalize=False)
    ind_info = sum(get_tree_splitwise_phylo_info(i) for i in (tree1, tree2))
    if normalize:
        return (ind_info - (2 * msi)) / ind_info
    return ind_info - (2 * msi)


def get_trees_shared_phylo_info(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return SPI score as sum of shared phylo info on matched splits.

    According to Martin Smith (2020):
    The SPI score measures how much the information shared between
    splits in a pair of trees narrows down the set of candidates
    that could be the ‘true’ tree, corresponding to the philosophy
    that phylogenetics seeks to reconstruct the single tree that
    accurately represents historical events.
    """
    biparts1 = list(tree1.iter_bipartitions())
    biparts2 = list(tree2.iter_bipartitions())
    arr, indices = _get_split_matching(biparts1, biparts2, "spi")
    spi = arr[indices].sum()
    if normalize:
        ind_info = sum(get_tree_splitwise_phylo_info(i) for i in (tree1, tree2))
        return spi / (ind_info / 2)
    return spi


def get_trees_shared_phylo_info_dist(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return difference in phylogenetic information betweern two trees.

    This is synonymous with the Phylogenetic Information Distance (pid)
    and thus synonymous with the function `get_treedist_pid`, although
    the latter has default of normalization=True.
    """
    spi = get_trees_shared_phylo_info(tree1, tree2, normalize=False)
    ind_info = sum(get_tree_splitwise_phylo_info(i) for i in (tree1, tree2))
    if normalize:
        return (ind_info - (2 * spi)) / ind_info
    return ind_info - (2 * spi)


def get_trees_shared_phylo_info_dist_from_biparts(
    biparts1: Sequence[Tuple[Tuple, Tuple]],
    biparts2: Sequence[Tuple[Tuple, Tuple]],
    normalize: bool = False,
) -> float:
    """Return difference in phylogenetic information betweern two trees.

    This is synonymous with the Phylogenetic Information Distance (pid)
    and thus synonymous with the function `get_treedist_pid`, although
    the latter has default of normalization=True.
    """
    arr, indices = _get_split_matching(biparts1, biparts2, "spi")
    spi = arr[indices].sum()
    info1 = sum(_get_split_phylo_info(i) for i in biparts1)
    info2 = sum(_get_split_phylo_info(i) for i in biparts2)
    ind_info = info1 + info2
    if normalize:
        return (ind_info - (2 * spi)) / ind_info
    return ind_info - (2 * spi)


def get_trees_mutual_clust_info(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return the mutual clustering information (cid).

    """
    biparts1 = list(tree1.iter_bipartitions())
    biparts2 = list(tree2.iter_bipartitions())
    arr, indices = _get_split_matching(biparts1, biparts2, "mci")
    mci = arr[indices].sum()
    if normalize:
        ind_info = sum(get_tree_splitwise_entropy(i) for i in (tree1, tree2))
        return mci / (ind_info / 2)
    return mci


@add_subpackage_method(TreeDistanceAPI)
def get_trees_mutual_clust_info_dist(
    tree1: ToyTree, tree2: ToyTree, normalize: bool = False,
) -> float:
    """Return the mutual clustering information distance (cid).

    """
    mci = get_trees_mutual_clust_info(tree1, tree2, normalize=False)
    ind_info = sum(get_tree_splitwise_entropy(i) for i in (tree1, tree2))
    if normalize:
        return (ind_info - (2 * mci)) / ind_info
    return ind_info - (2 * mci)


@add_subpackage_method(TreeDistanceAPI)
def get_trees_mutual_clust_info_dist_from_biparts(
    biparts1: Sequence[Tuple[Tuple, Tuple]],
    biparts2: Sequence[Tuple[Tuple, Tuple]],
    normalize: bool = False,
) -> float:
    """Return the mutual clustering information distance (cid).

    """
    arr, indices = _get_split_matching(biparts1, biparts2, "mci")
    mci = arr[indices].sum()
    info1 = sum(_get_split_entropy(b) for b in biparts1)
    info2 = sum(_get_split_entropy(b) for b in biparts2)
    ind_info = info1 + info2
    if normalize:
        return mci / (ind_info / 2)
    return mci


if __name__ == "__main__":

    import toytree

    # trees from ?TreeDist::SharedPhylogeneticInfo
    t1 = toytree.tree('((((a, b), c), d), (e, (f, (g, h))));')
    t2 = toytree.tree('(((a, b), (c, d)), ((e, f), (g, h)));')
    t3 = toytree.tree('((((h, b), c), d), (e, (f, (g, a))));')

    # t1 = toytree.tree('((A, B), ((C, (D, E)), (F, (G, (H, I)))));')
    # t2 = toytree.tree('((A, B), ((C, D, (E, I)), (F, (G, H))));')

    # phy info for a split:
    # print(_get_phylo_info(0, 0))


    # phy info for a tree (all splits): SplitWiseInfo(tree1) = 22.54
    print(get_tree_splitwise_phylo_info(t1)) # 22.54
    print(get_tree_splitwise_phylo_info(t2)) # 19.36

    print("MCI ", get_trees_mutual_clust_info(t1, t2)) # 3.03
    print("nMCI", get_trees_mutual_clust_info(t1, t2, True)) # 0.6908

    print("CID ", get_trees_mutual_clust_info_dist(t1, t2)) # 2.71
    print("nCID", get_trees_mutual_clust_info_dist(t1, t2, normalize=True)) # 0.3091

    print("MS ", get_trees_matching_split_dist(t1, t2)) # 6

    print("MSI ", get_trees_matching_split_info(t1, t2)) # 17.09
    print("nMSI", get_trees_matching_split_info(t1, t2, normalize=True)) # 0.815

    print("MSID ", get_trees_matching_split_info_dist(t1, t2)) # 7.72
    print("nMSID", get_trees_matching_split_info_dist(t1, t2, normalize=True)) # 0.185

    print("NYE ", get_trees_nye_similarity(t1, t2)) # 3.8
    print("nNYE ", get_trees_nye_similarity(t1, t2, normalize=True)) # 0.76

    print("NYED ", get_trees_nye_dist(t1, t2)) # 3.8
    print("nNYED ", get_trees_nye_dist(t1, t2, normalize=True)) # 0.76

    _report_matching(t1, t2, 'ms')

    # how similar are two trees? SharedPhylogeneticInfo(t1, t2) = 13.75
    # print(get_trees_phylo_info_shared(t1, t2), "shared")

    # how different are two trees? DifferentPhylogeneticInfo(t1, t2) = 14.40
    # print(get_trees_phylo_info_different(t1, t2, 1), "different")

    # _get_two_splits_shared_phylo_info()

    # PhylogeneticInfoDistance === DifferentPhylogeneticInfo
    # phy info diff for two trees: DifferentPhylogenetic Info
    # print(_get_split_matching(t1, t2, "mci"))
    # _visualize_matching(t1, t2)

    # split = set([range(60), range(60, 100)]) # 0.97 bits
    # split = set([range(50), range(50, 100)]) # 1 bits

    # split1 = set([range(60), range(60, 100)]) # 0.97 bits
    # split2 = set([range(0, 50), range(50, 100)]) # 1 bits
    # print(_get_entropy_distance(split1, split2))
    # print(_get_entropy_of_pair_split()  # 1.36 bits

    # t1 = toytree.rtree.unittree(6, seed=333)

    # for bipart in t1._iter_bipartitions():
    #     bip1 = ",".join(bipart[0])
    #     bip2 = ",".join(bipart[1])

    #     for dipart in t1._iter_bipartitions():
    #         dip1 = ",".join(dipart[0])
    #         dip2 = ",".join(dipart[1])

    #         e = _get_two_splits_entropy(bipart, dipart)
    #         print(
    #             f"{bip1}|{bip2} {_get_split_entropy(bipart):.2f} . "
    #             f"{dip1}|{dip2} {_get_split_entropy(dipart):.2f} . "
    #             f"{e[2]:.2f} {e[3]:.2f} {e[4]:.2f}"
            # )

    # t0 = toytree.tree('((a, b, c, d, e, f), (g, h, i, j));')
    # t1 = toytree.tree('((a, b, c, d, e, i, j), (g, h, f));')

    # b0 = t0._iter_bipartitions()
    # b1 = t1._iter_bipartitions()

    # for i in b0:
    #     for j in b1:
    #         print(i, j, _get_two_splits_matching_info(i, j))


#!/usr/bin/env python

"""A minimalist Python package for visualizing and studying evolution 
on trees.

Toytree's primary use if for visualizing and manipulating tree data
structures. It includes a number of additional subpackages for working
with trees as data, or data on trees. All subpackages make use only of
standard Python data science libs (e.g., numpy, scipy, pandas, numba) 
and does not include wrappers around any external tools.

"""

__version__ = "3.0.5"
__author__ = "Deren Eaton"

# core class objects
from toytree.core.node import Node
from toytree.core.tree import ToyTree
from toytree.core.multitree import MultiTree

# convenience functions
from toytree.io.src.treeio import tree
from toytree.io.src.mtreeio import mtree
from toytree.io.src.save import save

# toytree v3 supported subpackages
import toytree.rtree
import toytree.distance
import toytree.io
import toytree.mod
import toytree.color
import toytree.enum
import toytree.pcm
import toytree.network
import toytree.annotate
import toytree.data


# container trees... container

# start the logger at log_level WARNING
from toytree.utils.src.logger_setup import set_log_level
set_log_level("WARNING")



# stochastic_mapping.py

"""Stochastic character mapping.

This is a method of simulating a history for a discrete character
data set conditioned on the observed states at the tips ...
(and optionally at internal nodes?).

This method also needs a new plotting framework for displaying 
changes along edges of a tree using rectangles with linear-gradient
maps.

References
-----------
Bollback, J. P. (2006) Stochastic character mapping of discrete traits on phylogenies. BMC Bioinformatics, 7, 88.
Huelsenbeck, J. P., R. Neilsen, and J. P. Bollback (2003) Stochastic mapping of morphological characters. Systematic Biology, 52, 131-138.
"""

class StochasticMap:
    """Maps nodes to ndarrays of states and times?

    """



#!/usr/bin/env python

"""Annotation methods for adding node markers to tree drawings.

Examples
--------
...
"""

from typing import Tuple, Sequence, Mapping, Any, Union, TypeVar
import numpy as np
from loguru import logger

from toytree.core import ToyTree, Cartesian, Mark
from toytree.color import ToyColor
from toytree.core.apis import add_subpackage_method, AnnotationAPI
from toytree.style.src.validate_utils import substyle_dict_to_css_dict
from toytree.annotate.src.checks import get_last_toytree_mark, assert_tree_matches_mark
from toytree.drawing.src.mark_annotation import AnnotationRect, AnnotationMarker

from toytree.style.src.validate_data import (
    validate_colors,
    validate_numeric,
    validate_markers,
    validate_mask,
    validate_labels,
)
from toytree.style.src.validate_node_labels import validate_node_labels_style
from toytree.style.src.validate_nodes import validate_node_style


logger = logger.bind(name="toytree")
Color = TypeVar("Color", str, tuple, np.ndarray)
__all__ = [
    "add_node_markers",
    "add_node_labels",
    "add_node_bars",
    # "add_node_histograms",
    # "add_node_densigrams",
    # "add_node_pie_charts",  # see add_pie_markers.py
]


@add_subpackage_method(AnnotationAPI)
def add_node_markers(
    tree: ToyTree,
    axes: Cartesian,
    marker: Union[str, Sequence[str]] = "o",
    size: Union[int, Sequence[int]] = 8,
    color: Union[Color, Sequence[Color]] = None,
    opacity: Union[float, Sequence[float]] = None,
    mask: Union[np.ndarray, Tuple[int, int, int], bool, None] = None,
    xshift: int = 0,
    yshift: int = 0,
    style: Mapping[str, Any] = None,
) -> Mark:
    """Return a toyplot Mark of node markers added to a tree plot.

    This adds node markers to the last tree drawn on the Cartesian
    axes using the coordinates of plotted Nodes. The shape, size,
    color, and style of markers can be modified.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    marker: str or toyplot.marker.Marker or Sequence
        Marker shape, e.g., "o", "s", "^", "r2x1". See toyplot Markers.
    size: int or Sequence[int]
        Size of markers as single int or Sequence of ints, in px units.
    color: str, tuple, or array, or Sequence
        Color of markers as single color or Sequence of colors.
    opacity: float or Sequence[float]
        Opacity of markers (fill & stroke) as a float or Sequence
        of floats. Note: fill and stroke opacities can be set
        separately using the style dict, but only as single values.
    mask: np.array or None
        A boolean array of len nnodes or nnodes - 1 where True masks
        an node from being shown and False shows the edge. None shows
        all nodes. A tuple of 3 booleans can be entered as a shortcut
        to (show_tips, show_internal, show_root).
    style: dict
        Marker style dict. See `tree.style.node_style` for options.
    xshift: int
        Shift marker horizontally by px units (+=right, -=left).
    yshift: int
        Shift marker vertically by px units (+=down, -=up).

    Examples
    --------
    >>> tree = toytree.rtree.unittree(6, seed=123)
    >>> canvas, axes, m0 = tree.draw()
    >>> # add markers to all Nodes
    >>> m1 = tree.annotate.add_node_markers(
    >>>     axes,
    >>>     marker='s',
    >>>     size=9,
    >>>     color='red',
    >>>     style={'stroke': 'white', 'stroke-width': 2.5}
    >>> )
    >>>
    >>> # add markers to only a few Nodes
    >>> m2 = tree.annotate.add_node_markers(
    >>>     axes, marker=">", size=20, mask=tree.get_node_mask(9)
    >>> )
    """
    # get mark for coordinates on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)
    coords = mark.ntable

    # get mask and apply to all other styles below
    mask = validate_mask(tree, style={"node_mask": mask})

    # update node_style setting
    style = {} if style is None else style
    style = validate_node_style(tree, style=None, **style)
    style = substyle_dict_to_css_dict(style.__dict__)

    # update node colors setting; sets to None if only one color.
    colors, fill_color = validate_colors(
        tree, key="color", size=tree.nnodes, style={"color": color})

    # if fill_color then set to node_style.fill since colors = None
    if colors is None:
        if fill_color:
            style["fill"] = ToyColor(fill_color)  # overrides node_style.fill
        else:
            pass  # node_style.fill overrides
    else:
        colors = colors[mask]
        style.pop("fill")

    # validate others and trim to mask
    markers = validate_markers(
        tree, key="marker", size=tree.nnodes, style={"marker": marker})[mask]
    sizes = validate_numeric(
        tree, key="size", size=tree.nnodes, style={"size": size})[mask]
    opacity = validate_numeric(
        tree, key="opacity", size=tree.nnodes, style={"opacity": opacity})[mask]

    # if all marker opacities are the same then set to 1 and use style.
    # this simplifies the CSS and makes things faster, but is otherwise
    # not necessary. EDIT: commented out for now until `concat_tyle_fix_color`
    # used in render_annotations.py can handle fill-opacity w/o fill.
    # if len(set(opacity)) == 1:
    #     style["fill-opacity"] = opacity[0]
    #     opacity = None
    mark = AnnotationMarker(
        ntable=coords[mask],
        xshift=xshift,
        yshift=yshift,
        sizes=sizes,
        colors=colors,
        opacity=opacity,
        shapes=markers,
        style=style,
    )
    axes.add_mark(mark)
    return mark

    # # apply mask to Node coordinates
    # # TODO: alternative strategy here
    # coords = coords[mask, :]
    # if xshift or yshift:
    #     # Note: if later annotations change the projection this will be off
    #     axes._finalize()
    #     if xshift:
    #         origin, xshifted = axes._x_projection.inverse([0, xshift])
    #         x_shift_projected = xshifted - origin
    #         coords[:, 0] += x_shift_projected
    #     if yshift:
    #         origin, yshifted = axes._y_projection.inverse([0, yshift])
    #         y_shift_projected = yshifted - origin
    #         coords[:, 1] += y_shift_projected
    #     axes._finalized = None

    # # plot edge markers as scatterplot markers
    # mark = axes.scatterplot(
    #     coords[:, 0],
    #     coords[:, 1],
    #     color=node_colors,
    #     size=sizes,
    #     marker=markers,
    #     mstyle=style,
    #     opacity=opacity,
    # )
    # return mark


@add_subpackage_method(AnnotationAPI)
def add_node_labels(
    tree: ToyTree,
    axes: Cartesian,
    labels: Union[str, Sequence[str]] = "idx",
    color: Union[Color, Sequence[Color]] = None,
    opacity: Union[float, Sequence[float]] = 1.0,
    font_size: Union[int, None] = 12,
    angle: Union[int, Sequence[int]] = 0,
    mask: Union[np.ndarray, Tuple[int, int, int], None] = None,
    xshift: int = 0,
    yshift: int = 0,
    style: Mapping[str, Any] = None,
) -> Mark:
    """Return a toyplot Mark of node labels added to a tree drawing.

    This adds node labels to the last tree drawn on the Cartesian axes.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    labels: str or Sequence[str]
        A sequence of labels as strings in Node idx traversal order.
        The length must be equal to nnodes or nnodes - 1. Use '' to
        not add a label to some edges. You can also enter a single str
        as the name of a feature on the tree to extract data as labels,
        e.g., "idx" will show int idx labels for nodes.
    color: str, tuple, array or Sequence
        A single color or Sequence of colors for node labels.
    opacity: float or Sequence[float]
        A single opacity or Sequence of opacities for node labels.
    font_size: float
        Font size in px. Overrides 'font-size' setting in style dict.
    angle: int or Sequence[int]
        A single angle applied to all labels, or Sequence of angles.
    mask: np.array or None
        A boolean array of len nnodes or nnodes - 1 where True masks
        an node from being shown and False shows the node. None or False
        shows all nodes. A tuple of 3 booleans can be entered as a
        shortcut to (show_tips, show_internal, show_root).
    xshift: int
        Shift label horizontally by px units (+=right, -=left).
    yshift: int
        Shift label vertically by px units (+=down, -=up).
    style: dict
        Style dict. See `tree.style.node_labels_style` for options.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(6, seed=123)
    >>> canvas, axes, m0 = tree.draw()
    >>> m1 = tree.annotate.add_node_labels(
    >>>     axes,
    >>>     labels=tree.get_node_data("idx"),
    >>>     color='blue',
    >>>     style={'font-size': 16, 'baseline-shift': 8}
    >>> )
    """
    # get mark for coordinates on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # mask some edges
    mask = validate_mask(tree, style={"node_mask": mask})
    coords = mark.ntable[mask]

    # check length and type of labels
    labels = validate_labels(
        tree, key="labels", size=tree.nnodes, style={"labels": labels})[mask]

    # set styles on top of defaults
    style = {} if style is None else style
    style = validate_node_labels_style(tree, style=style, **style)
    style = substyle_dict_to_css_dict(style.__dict__)

    # override font size
    if font_size:
        style["font-size"] = font_size

    # update node colors setting; sets to None if only one color.
    node_colors, fill_color = validate_colors(
        tree, key="color", size=tree.nnodes, style={"color": color})

    # if fill_color then set to node_style.fill since node_colors = None
    if node_colors is None:
        if fill_color:
            style["fill"] = ToyColor(fill_color)  # overrides node_style.fill
        else:
            pass  # node_style.fill overrides
    else:
        node_colors = node_colors[mask]
        style.pop("fill")

    # ...
    opacity = validate_numeric(
        tree, key="opacity", size=tree.nnodes, style={"opacity": opacity})[mask]
    angle = validate_numeric(
        tree, key="angle", size=tree.nnodes, style={"angle": angle})[mask]

    # expand xshift,yshift args as anchor_shift,baseline_shift
    style['-toyplot-anchor-shift'] += xshift
    style['baseline-shift'] -= yshift

    # add text at Node positions + half length of dists.
    mark = axes.text(
        coords[:, 0],
        coords[:, 1],
        labels,
        color=node_colors,
        opacity=opacity,
        angle=angle,
        style=style,
        annotation=True,
    )
    return mark


@add_subpackage_method(AnnotationAPI)
def add_node_bars(
    tree: ToyTree,
    axes: Cartesian,
    bar_min: Union[int, Sequence[int], str],
    bar_max: Union[int, Sequence[int], str],
    # Note: what about CI(5-95) features? suggest parsing... enter only as min?
    color: Union[Color, Sequence[Color]] = None,
    mask: Union[np.ndarray, bool, Tuple, None] = (0, 1, 1),
    size: Union[float, Sequence[float]] = 0.5,
    opacity: Union[float, Sequence[float]] = 0.7,
    xshift: int = 0,
    yshift: int = 0,
    style: Mapping[str, Any] = None,
    z_index: int = 0,
) -> Mark:
    """Returns a toyplot marker to add bars at Nodes.

    This is commonly used to display confidence intervals on node ages.
    Bars are rectangles on which styles can be set.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    bar_min: float, Series of float, or str
        Minimum height of the bar.
    bar_max: float, Series of float, or str
        Maximum height of the bar.
    color: str, tuple, array or Sequence
        A single color or Sequence of colors for node labels.
    mask: np.array or None
        A boolean array of len nnodes or nnodes - 1 where True masks
        an node from being shown and False shows the node. None or False
        shows all nodes. A tuple of 3 booleans can be entered as a
        shortcut to (show_tips, show_internal, show_root).
    xshift: int
        Shift label horizontally by px units (+=right, -=left).
    yshift: int
        Shift label vertically by px units (+=down, -=up).
    style: dict
        Style dict. See `tree.style.node_labels_style` for options.
    z_index: int
        Index of annotation (default=0). Lower index makes the bars
        appears behind other marks (e.g., tree or other annotations).

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10, treeheight=1e5)
    >>> c, a, m = tree.draw()
    >>> node_height = tree.get_node_data("height").values
    >>> tree.annotate.add_node_bars(
    >>>     axes=a,
    >>>     bar_min=node_height * 0.5,
    >>>     bar_max=node_height * 1.5,
    >>>     size=0.33, z_index=-1, color='purple', opacity=0.4,
    >>> )
    """
    # get mark for coordinates of Nodes on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # mask some edges
    mask = validate_mask(tree, style={"node_mask": mask})

    # get Node coordinates
    coords = mark.ntable.copy()[mask]

    # update node_style setting
    base_style = {"stroke": None, "fill-opacity": 0.5}
    if style is not None:
        base_style.update(style)
    style = validate_node_style(tree, style=None, **base_style)
    style = substyle_dict_to_css_dict(style.__dict__)

    # update node colors setting; sets to None if only one color.
    opacity = validate_numeric(
        tree, key="opacity", size=tree.nnodes, style={"opacity": opacity})[mask]
    colors, fill_color = validate_colors(
        tree, key="color", size=tree.nnodes, style={"color": color})

    # if fill_color then set to node_style.fill since node_colors = None
    if colors is None:
        if fill_color:
            style["fill"] = ToyColor(fill_color)  # overrides node_style.fill
        else:
            pass  # node_style.fill overrides
    else:
        colors = colors[mask]
        style.pop("fill")

    # check values for positions
    bar_min = validate_numeric(
        tree, key="bar_min", size=tree.nnodes, style={"bar_min": bar_min})[mask]
    bar_max = validate_numeric(
        tree, key="bar_max", size=tree.nnodes, style={"bar_max": bar_max})[mask]
    sizes = validate_numeric(
        tree, key="size", size=tree.nnodes, style={"size": size})[mask]

    # orient rectangles for the tree layout
    if mark.layout == "r":
        xtable = np.column_stack([-bar_max, -bar_min])
        ytable = np.column_stack([coords[:, 1] - sizes / 2., coords[:, 1] + sizes / 2.])
    elif mark.layout == "l":
        xtable = np.column_stack([bar_min, bar_max])
        ytable = np.column_stack([coords[:, 1] - sizes / 2., coords[:, 1] + sizes / 2.])
    elif mark.layout == "d":
        xtable = np.column_stack([coords[:, 0] - sizes / 2., coords[:, 0] + sizes / 2.])
        ytable = np.column_stack([bar_min, bar_max])
    elif mark.layout == "u":
        xtable = np.column_stack([coords[:, 0] - sizes / 2., coords[:, 0] + sizes / 2.])
        ytable = np.column_stack([-bar_max, -bar_min])
    else:
        raise NotImplementedError("TODO")

    # build the Mark
    mark = AnnotationRect(
        ntable=coords,
        xtable=xtable,
        ytable=ytable,
        colors=colors,
        opacity=opacity,
        xshift=xshift,
        yshift=yshift,
        style=style,
    )

    # extend the domain to ensure bars fit
    mark._annotation = True

    # z-index: option to set markers to appear UNDER the tree Mark
    axes._scenegraph._relationships['render']._targets[axes].insert(z_index, mark)
    axes._scenegraph._relationships['map']._targets[axes.x].insert(z_index, mark)
    axes._scenegraph._relationships['map']._targets[axes.y].insert(z_index, mark)
    return mark


# def add_clade_box(
#     tree: ToyTree,
#     axes: Cartesian,
# ):
#     pass


if __name__ == "__main__":

    import toytree

    # base tree drawing
    tree = toytree.rtree.unittree(12)
    c, a, m = tree.draw(layout='r', scale_bar=True, node_sizes=5, width=400)

    m0 = tree.annotate.add_node_markers(a, color="idx", yshift=-15, opacity=0.3)
    # m1 = tree.annotate.add_node_labels(a, font_size=20, yshift=-15)
    # m2 = tree.annotate.add_node_bars(
    #     a,
    #     bar_min=tree.get_node_data("height").values * 0.8,
    #     bar_max=tree.get_node_data("height").values * 2,
    #     size=0.33,
    #     z_index=-1,
    #     color='purple',
    #     opacity=1.0,
    #     style={"fill-opacity": 0.3, "stroke": None},
    #     # yshift=15,
    #     # xshift=15,
    # )
    toytree.utils.show(c, tmpdir="~")



#!/usr/bin/env python

"""ColorKit class for parsing and converting between color formats.

This is used to parse a SINGLE color input. This is an internal func
not intended for users. Instead, users should use the ToyColor object,
which uses ColorKit as its parser.
"""

from typing import Union, Tuple
from copy import copy
import numpy as np
from loguru import logger
import toyplot.color
from toytree.utils import ToyColorError

logger = logger.bind(name="toytree")
ColorType = Union[str, np.ndarray, Tuple[float, float, float, float]]


class ColorKit:
    """Flexible color parser class to get css or rgba tuples.

    Color formats
    -------------
    - rgba = (r:float, g:float, b:float, a:float)
    - rgb = (r:float, g:float, b:float, a:float)
    - hex = "#000000"
    - css = "cornflowerblue"
    - array = np.array((1.0, 0.0, 0.0, 0.5), dtype=toyplot.color.dtype)
    """
    def __init__(self, color: ColorType):

        # store color as css, rgba, and ndarray
        self._css: str = None
        self._rgba: Tuple[float, float, float, float] = None
        self._array: np.ndarray = None
        self._parse_color(color)

        self._style = {
            "float": "left",
            "width": "16px",
            "height": "16px",
            "margin-right": "5px",
            "background-color": self.css,
            "border-radius": "50%",
        }

    @property
    def css(self) -> str:
        """Return CSS string as 'rgba(R%,G%,B%,A%)'."""
        return self._css

    @property
    def array(self) -> np.ndarray:
        """Return numpy array with complex (r,g,b,a) dtype."""
        return self._array

    @property
    def rgba(self) -> Tuple[float, float, float, float]:
        """Return tuple with float for (r,g,b,a) percentages."""
        return self._rgba

    @property
    def rgb(self) -> Tuple[float, float, float]:
        """Return tuple with float for (r,g,b) percentages, no a."""
        return tuple(float(self.array[i]) for i in 'rgb')

    def set_opacity(self, value: float) -> None:
        """Set opacity to a new value."""
        rgba = list(self._rgba)
        rgba[3] = value
        self._parse_color(tuple(rgba))
        return self

    def get_style_str(self, stroke: bool = False, opacity: float = None) -> str:
        """Return SVG style string with fill/stroke and opacity separate.

        If opacity is set to False then no opacity is printed. This is
        used when the parent class has an opacity that should be
        inherited.
        """
        # choose stroke or fille
        coltype = "stroke" if stroke else "fill"
        # get rbg from rgba
        rgb_css = f"rgb({self.css[5:].rsplit(',', 1)[0]})"
        # store rgb color as string
        color = f"{coltype}:{rgb_css}"
        # if opacity is None then get value from rgba
        if opacity is None:
            opacity = f";{coltype}-opacity:{self.rgba[-1]}"
        # if opacity if False then do not set.
        elif opacity is False:
            opacity = ""
        # if opacity if a value then use it.
        else:
            opacity = f";{coltype}-opacity:{opacity}"
        # return style string
        return color + opacity

    def _parse_color(self, color: ColorType) -> None:
        """Parse input color to three stored formats."""
        # input is a css string (parse input based on type

        # logger.error(color)
        if isinstance(color, str):
            if color == "none":
                self._rgba = (0, 0, 0, 0)
                self._array = toyplot.color.rgba(*self._rgba)
                self._css = toyplot.color.to_css(self._array)
            else:
                self._array = toyplot.color.css(color)
                if self._array is None:
                    raise ToyColorError(f"CSS color '{color}' not recognized")
                self._css = toyplot.color.to_css(self._array)
                self._rgba = tuple(float(self._array[i]) for i in 'rgba')

        # input is an ndarray (e.g., toyplot.color ndarray, 1-d 4 floats)
        elif isinstance(color, np.ndarray) and color.dtype == toyplot.color.dtype:
            if color.size > 1:
                raise ToyColorError(
                    "Cannot parse a multi-color array to a ToyColor. "
                    "Use `toytree.color.ToyColor.color_expander()`\n"
                    "to convert it to a List[ToyColor]."
                )
            self._array = color.copy()
            self._css = toyplot.color.to_css(color)
            if not self._css:
                raise ToyColorError(f"color array {color} not recognized")
            self._rgba = tuple(float(self._array[i]) for i in 'rgba')

        elif isinstance(color, tuple) and len(color) == 4:
            self._rgba = color
            self._array = toyplot.color.rgba(*self._rgba)
            self._css = toyplot.color.to_css(self._array)

        elif isinstance(color, tuple) and len(color) == 3:
            self._rgba = color + (1.0,)
            self._array = toyplot.color.rgba(*self._rgba)
            self._css = toyplot.color.to_css(self._array)

        elif isinstance(color, ColorKit):
            self._array = copy(color._array)
            self._css = copy(color._css)
            self._rgba = copy(color._rgba)

        elif isinstance(color, np.void):
            self._rgba = copy(color.item())
            self._array = toyplot.color.rgba(*self._rgba)
            self._css = toyplot.color.to_css(self._array)

        elif isinstance(color, float):
            if np.isnan(color):
                self._rgba = (0, 0, 0, 0)
                self._array = toyplot.color.rgba(*self._rgba)
                self._css = toyplot.color.to_css(self._array)
            else:
                raise ToyColorError(
                    f"Color arg '{color}' not recognized as a valid color input.")

        elif color is None:
            self._rgba = (0, 0, 0, 0)
            self._array = toyplot.color.rgba(*self._rgba)
            self._css = toyplot.color.to_css(self._array)

        else:
            raise ToyColorError(
                f"Color arg '{color}' not recognized as a valid color input.")

    # Colorkit doesn't need a repr, but this one was pretty nice, so 
    # I'll leave it here for now. But I commented it b/c I was encountering
    # a circular import error and didn't feel like fixing it.
    # def _repr_html_(self):
    #     """Show color as a div in jupyter notebooks"""
    #     # create a root dom element
    #     root_xml = xml.Element(
    #         "div",
    #         style="overflow:hidden; height:auto; display:table",
    #         attrib={"class": "toytree-ToyColor"},
    #     )
    #     _ = xml.SubElement(root_xml, "div", style=concat_style_to_str(self._style))
    #     _ = xml.SubElement(root_xml, "text").text = (
    #         "Color({:.2f}, {:.2f}, {:.2f}, {:.2f})"
    #         .format(*self.rgba)
    #     )

    #     # convert deom to a simple html script
    #     html = xml.tostring(root_xml, encoding="unicode", method="html")
    #     return html


if __name__ == "__main__":

    COLORS = [
        "RED",
        "red",
        "#000000",
        "rgba(100.0%,50.0%,25.0%,0.500)",
        "rgb(100%,50%,25%)",
        (1.0, 0.5, 0.25, 0.5),
        (1.0, 0.5, 0.25),
        np.array((1.0, 0.5, 0.25, 0.5), dtype=toyplot.color.dtype),
        ColorKit("red"),
    ]
    for i in COLORS:
        c = ColorKit(i)
        print(c.css, c.rgba, c.array, c.get_style_str(), "\t | original =", i)

    # set opacity
    print(c.rgba)
    c.set_opacity(0.5)
    print(c.rgba)


#!/usr/bin/env python

"""Calculate parsimony score of a tree topology given data.

TODO
----
- support DNA IUPAC models where ambiguous states are automatically
expanded.
- support non DNA (arbitrary multistate including binary) data.
- vectorize to be faster.

JIT-compiled version
---------------------
- jit supports sets now, but they are being deprecated to introduce
a new numba typed.set, but which does not yet exist. So until it does
its not really worth spending time on...
- https://numba.pydata.org/numba-doc/dev/reference/pysupported.html#set

References
----------
- Xia, Xuhua. 2018. “Maximum Parsimony Method in Phylogenetics.”
  In Bioinformatics and the Cell: Modern Computational Approaches in
  Genomics, Proteomics and Transcriptomics, edited by Xuhua Xia,
  327–41. Cham: Springer International Publishing.
  https://doi.org/10.1007/978-3-319-90684-3_14.
- Fitch, Walter M. 1971. “Toward Defining the Course of Evolution:
  Minimum Change for a Specific Tree Topology.” Systematic Biology 20
  (4): 406–16. https://doi.org/10.1093/sysbio/20.4.406.

- Sankoff (1975)
- Felsenstein (2004)
- BioPython
- https://telliott99.blogspot.com/search/label/maximum%20likelihood
- https://telliott99.blogspot.com/2010/03/fitch-and-sankoff-algorithms-for.html
"""

from typing import TypeVar, Optional
from numpy.typing import ArrayLike

ToyTree = TypeVar("ToyTree")


def get_parsimony_score(
    tree: ToyTree,
    data: ArrayLike,
    weights: Optional[ArrayLike],
    data_as_dna: bool = True,
):
    """Return the parsimony score of a tree given a data matrix.

    The parsimony score is calculated by performing a post-order
    traversal on a rooted tree (if unrooted it will be arbitrarily
    rooted, which has no effect on score) and setting the state of
    internal nodes to be the set intersection of states at its
    descendant Nodes. If no intersection exists, the union is instead
    set and this counts as a 'change' event. If a weights matrix is
    entered then this indicates the contribution of each type of
    change towards the 'score' (Sankoff parsimony), else each change
    counts as 1 (Fitch parsimony). If multiple states exist in the
    data matrix the sum of scores of each state is returned.

    Parameters
    ----------
    tree: ToyTree
        A tree on which to calculate the parsimony score.
    data: ArrayLike
        A data matrix of shape (ntips, ntraits) containing discrete
        values for one or more traits for each tip Node in idxorder.
    weights: None or ArrayLike
        The weights matrix must be square. If it is a dataframe then
        the row and column names will be used, else they should be
        ordered by state values alphanumerically (e.g., 0, 1, 2, 3 or
        'A', 'C', 'G', 'T'). The diagonal must be zero, and
        off-diagonal as float or int types.
    data_as_dna: bool
        If True then data values that are string types representing
        DNA IUPAC ambiguity codes (e.g., RWMYSK) will be expanded to a
        set of the two bases that they represent (e.g., W -> {A, T}).

    Note
    ----
    This function uses numba jit-compilation for speed improvements.

    References
    ----------
    - ...

    Examples
    --------
    >>> ...
    """
    # get tree as an array of idxs in postorder traversal
    traversal_order = np.array([i.idx for i in tree.traverse("postorder")])

    # get data as a array of ints after expanding ambiguities
    if data_as_dna:
        pass
    else:
        pass

    # calculate sum of scores and return
    score = jit_compiled_fitch(
        tree=traversal_order,
        data=int_matrix,
        weights=weights_matrix,
    )
    return score


class Parsimony:
    """Infer a phylogenetic tree by Maximum Parsimony.

    Note
    ----
    The maximum parsimony algorithm does not infer rooted trees, i.e.,
    topologies re-rooted at any arbitrary edge will yield the same
    parsimony score.

    Examples
    ---------
    >>> toytree.set_log_level("DEBUG")  # print verbose information
    >>> data = ...
    >>> starting_tree = ... distance tree.
    >>> tool = Parsimony(data)
    >>> tool.get_score(tree=tree, )
    """
    def __init__(self, data: ...):
        self.data = data

    def get_score(self, tree):
        """Return parsimony score"""


    def _fitch_algorithm(self):
        """Implement the fitch algorithm.

        """

    def _sankoff_algorithm(self):
        """Implement the Sankoff algorithm.

        The strength of the Sankoff algorithm is that it allows a
        variety of cost matrices to be used. This is in principal
        closer to ML, where we would define a substitution model.
        Here the cost matrix is not inferred, but a priori defined
        by the user.

        """

    def _tree_move(self, method: str):
        """Return a ToyTree that is one 'move' from the current tree.

        >>> NNITreeSearcher(scorer)
        """
        pass


class Fitch:
    """Implementation of the Fitch parsimony algorithm."""
    pass


def _validate_against_bio():
    """Compare against Bio implementation for validation"""
    pass


if __name__ == "__main__":

    # test parsimony score and inference against Bio
    import toytree
    tree = toytree.rtree.unittree(10, treeheight=1000, seed=123)
    # data = tree.pcm.simulate_discrete_markov_data("seqs", states=4, reps=10)




#!/usr/bin/env python

"""Utilities for visualizing TreeSequences from tskit or SLiM.

The :class:`ToyTreeSequence` class holds the tskit.trees.TreeSequence 
object in its `.tree_sequence` attribute. It is primarily used to 
generate drawings of one or more trees with the option to display 
mutations, MutationTypes, and a chromosome structure. These latter 
options are mostly for SLiM simulated TreeSequences.

Examples
--------
>>> tts = toytree.utils.tree_sequence.ToyTreeSequence(tree_sequence)
>>> tts.draw_tree(idx=0)
>>> tts.draw()
"""

from typing import List, Tuple, Optional, Iterable
from dataclasses import dataclass
import toyplot
from toytree.utils.src.scrollable_canvas import ScrollableCanvas
import toytree


@dataclass
class Box:
    """Delimited boxes used in TreeSequenceDrawing."""
    left: float
    right: float
    top: float
    bottom: float


@dataclass
class MultiDrawing:
    trees: List['ToyTree']
    breakpoints: List[float]
    width: float
    height: float
    padding: float = 20
    margin: Tuple[float,float,float,float] = (50, 50, 50, 50)

    def __post_init__(self):
        self.height = self.height if self.height is not None else 325
        self.width = max(300, (
            self.width if self.width is not None else 
            15 * self.trees[0].ntips * len(self.trees)
        ))


class ToyTreeSequenceDrawing(MultiDrawing):
    """Return a drawing of multiple trees from a tree sequence.

    A ToyTreeSequence...

    Parameters
    ----------
    """
    def __init__(
        self,
        trees: List['ToyTree'],
        breakpoints: Iterable[int],
        width: Optional[int]=None,
        height: Optional[int]=None,
        padding: float=20,
        margin: Tuple[float,float,float,float]=(50,50,50,50),
        colormap: List['color']=None,
        scrollable: bool=True,
        axes: Optional['toyplot.coordinates.Cartesian']=None,
        **kwargs,
    ):
        super().__init__(trees, breakpoints, width, height, padding, margin)
        self.canvas = self.get_canvas(scrollable, axes)
        self.axes = self.get_axes(axes)
        self.marks = []

        # colormap
        self.icolors = toytree.color.color_cycler(colormap)

        # positioning
        self.xprop_space = 0.2
        self.xspace = max(2., self.xprop_space * self.trees[0].ntips - 1)
        self.xtree = self.xspace + self.trees[0].ntips - 1
        self.xmax = self.xtree * len(self.trees)
        self.ymax = max(tre.treenode.height for tre in self.trees)

        # store
        self.boxes = {}
        self.colors = {}

        # run funcs
        self.get_position_bar_marks()
        self.get_polygon_container_marks()
        self.get_tree_marks(kwargs)
        self.set_axes_style()

    def __repr__(self):
        return "<TreeSequenceDrawing>"

    def get_canvas(self, scrollable, axes):
        """Return a toyplot.Canvas with option for width overflow."""
        # if an existing axes was passed then use its canvas
        if axes is not None:
            return None
        # create a new canvas
        if scrollable:
            canvas = ScrollableCanvas(width=self.width, height=self.height)
        else:
            canvas = toyplot.Canvas(width=self.width, height=self.height)
        return canvas

    def get_axes(self, axes):
        """Get toyplot axes, or use input one. 
        Makes a 'share' axes so that ticks appear on top of plot.
        """
        if axes is not None:
            return axes
        axes = self.canvas.cartesian(margin=self.margin, padding=self.padding)
        axes2 = axes.share('y')
        axes.x.show = False
        return axes2


    def get_position_bar_marks(self):
        """Get breakpoint bars on top spanning horizontal axis.
        """
        for idx, pos in enumerate(self.breakpoints[:-1]):

            # top bar starts with \ and ends with /
            pos0 = self.xmax * (self.breakpoints[idx] / max(self.breakpoints))
            pos1 = self.xmax * (self.breakpoints[idx + 1] / max(self.breakpoints))
            if pos == self.breakpoints[-2]:
                pos1 += self.xspace

            box = Box(
                left=pos0,
                right=pos1,
                top=self.ymax + self.ymax * 0.25,
                bottom=self.ymax + self.ymax * 0.2,
            )
            self.boxes[idx] = box
            self.colors[idx] = next(self.icolors)

            self.axes.fill(
                [box.left, box.right],
                [box.bottom, box.bottom],
                [box.top, box.top],
                style={
                    "fill": self.colors[idx],
                    "fill-opacity": 0.85,
                    'stroke': 'white',
                    'stroke-opacity': 0.5,
                },
            )

    def get_polygon_container_marks(self):
        """
        Get polygon fill marks to surround trees.
        """
        for idx, _ in enumerate(self.trees):
            box = self.boxes[idx]
            tbox = Box(
                left=self.xspace / 2 + (idx * self.xtree),
                right=self.xspace / 2 + ((idx + 1) * self.xtree),
                top=self.ymax,
                bottom=0,
            )

            # hover pop-up
            title = "\n".join([
                f"idx: {idx}",
                f"interval: ({self.breakpoints[idx]:.0f} - {self.breakpoints[idx + 1]:.0f})",
                f"tmrca: {round(self.trees[idx].treenode.height, 2)}",
                #f"mutations: 0", # todo
                #f"alleles: 1", # todo
            ])

            # polygon style
            pstyle = {
                "fill": self.colors[idx],
                'fill-opacity': 0.25,
                'stroke': 'white',
                'stroke-opacity': 0.5,
            }

            # draw polygons
            if box.left < tbox.right:
                self.axes.fill(
                    [box.left,   tbox.left,  tbox.left,   tbox.right,  tbox.right, box.right],
                    [box.bottom, tbox.top,   tbox.bottom, tbox.bottom, tbox.top,   box.bottom],
                    [box.bottom, box.bottom, box.bottom,  box.bottom,  box.bottom, box.bottom],
                    annotation=False,
                    title=title,
                    style=pstyle,
                )
            else:
                self.axes.fill(
                    [tbox.left,   box.left,    box.right,   tbox.right,],
                    [tbox.bottom, tbox.bottom, tbox.bottom, tbox.bottom],
                    [tbox.top,    box.bottom,  box.bottom,  tbox.top,  ],
                    annotation=False,
                    title=title,
                    style=pstyle,
                )

    def get_tree_marks(self, kwargs):
        """Get ToyTree marks.
        """
        for idx, tree in enumerate(self.trees):
            base_style = {
                'xbaseline': self.xspace + (idx * self.xtree),
                'layout': 'd',
                'scale_bar': True,
                'tip_labels_style': {"font-size": "10px"}
            }
            base_style.update(kwargs)
            _, _, mark = tree.draw(axes=self.axes, **base_style)
            # mark.annotation = True
            self.marks.append(mark)

            # build mutation marker info
            xpos = []
            ypos = []
            titles = []
            colors = []
            for mut in tree.mutations:

                # get node id and time using the 'tsidx' (tskit node id)
                node = tree.tsidx_dict[mut.node]
                time = mut.time

                # skip if on root edge (shared by all)
                if node.is_root():
                    continue

                # project mutation points to proper layout type
                xpos.append(node._x + mark.xbaseline)
                ypos.append(time)

                # store color and title for this point
                try:
                    mtype = int(mut.metadata['mutation_list'][0]['mutation_type'])
                except Exception:
                    mtype = 0
                    
                color = toytree.color.COLORS1[mtype]
                colors.append(color)
                title = (
                    f"id: {mut.id}\n"
                    f"pos: {mut.site:.0f}\n"
                    f"time: {mut.time:.0f}\n"
                    f"mtype: {mtype}")
                titles.append(title)

            # update mutation style dict
            mstyle = {"stroke": "black"}

            # draw the mutations
            if xpos:
                mark = self.axes.scatterplot(
                    xpos, ypos,
                    marker='o',
                    size=7,
                    color=colors,
                    mstyle=mstyle,
                    title=titles,
                )
                mark.annotation = True
                self.marks.append(mark)

    def set_axes_style(self):
        """Adds ..."""
        # get tick marks within domain 
        locator = toyplot.locator.Extended(only_inside=True)
        ticks = locator.ticks(0, max([i.top for i in self.boxes.values()]))
        self.axes.y.ticks.locator = toyplot.locator.Explicit(*ticks)

        # import numpy as np
        # locator = toyplot.locator.Extended(count=10, only_inside=True)
        # ticks = locator.ticks(0, self.xmax)
        # labels = locator.ticks(min(self.breakpoints), max(self.breakpoints))
        # self.axes.x.ticks.locator = toyplot.locator.Explicit(
        #     ticks[0], labels[1])


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Testing simple alternative validation

"""

from typing import Any, Dict, TypeVar
import numpy as np

from toytree.color import ToyColor
from toytree.style.src.style_base import TreeStyle, EdgeStyle, EdgeAlignStyle

ToyTree = TypeVar("ToyTree")
Color = TypeVar("Color", str, tuple, np.ndarray)

__all__ = [
    "validate_edge_style",
    "validate_edge_align_style",
]


# STILL USED!
def validate_edge_style(
    tree: ToyTree,
    style: TreeStyle = None,
    **kwargs,
) -> Dict[str, Any]:
    """Return Dict with css style keys and ensure fill and stroke values.

    Parameters
    ----------
    base: TreeStyle or None
        A base tree style to modify. If None then new NodeLabelStyle is used.
    style: Mapping
        A dict of style changes.
    """
    # get user value or style base value
    if style is None:
        style = EdgeStyle()

    # update default style with user style args
    for key, val in kwargs.items():
        style[key] = val

    # convert stroke to ToyColor
    style.stroke = ToyColor(style.stroke)

    # optionally convert fill and stroke to css string: ??
    # "fill:{rgb};fill-opacity:{o};stroke:{rgb};stroke-opacity:{o}"
    # if serialize:
    #     style['stroke'] = style['stroke'].css
    #     if not style.get("stroke-opacity"):
    #         style['stroke-opacity'] = 1.0
    return style


# STILL USED!
def validate_edge_align_style(
    tree: ToyTree,
    style: TreeStyle = None,
    **kwargs,
) -> Dict[str, Any]:
    """Return Dict with css style keys and ensure fill and stroke values.

    Parameters
    ----------
    base: TreeStyle or None
        A base tree style to modify. If None then new NodeLabelStyle is used.
    style: Mapping
        A dict of style changes.
    """
    # get user value or style base value
    if style is None:
        style = EdgeAlignStyle()

    # update default style with user style args
    for key, val in kwargs.items():
        style[key] = val

    # convert stroke to ToyColor
    style.stroke = ToyColor(style.stroke)

    # check dasharray type
    assert isinstance(style.stroke_dasharray, str), (
        "edge_align_style 'stroke-dasharray' value must be a str "
        "e.g., '2,2'")

    # optionally convert fill and stroke to css string: ??
    # "fill:{rgb};fill-opacity:{o};stroke:{rgb};stroke-opacity:{o}"
    # if serialize:
    #     style['stroke'] = style['stroke'].css
    #     if not style.get("stroke-opacity"):
    #         style['stroke-opacity'] = 1.0
    return style


if __name__ == "__main__":

    pass



#!/usr/bin/env python

"""toytree I/O data parsing utilities.

"""

from toytree.io.src.newick import parse_newick_string, parse_newick_string_custom
from toytree.io.src.treeio import tree
from toytree.io.src.mtreeio import mtree
from toytree.io.src.writer import write



#!/usr/bin/env python

"""A custom toyplot mark Class built from the toyplot Mark constructor.

The purpose of a Mark is to set the `domain` and `extents` of an
mark, so that it can be fit on coordinate axes. This object stores
this info along with the input Node coordinates and style args that
have already been checked for validity.
"""

from typing import List, Tuple, Union, Sequence
import toyplot
import numpy as np
from loguru import logger
from toyplot.mark import Mark
import toyplot.text

logger = logger.bind(name="toytree")


# Note: see toytree/drawing/src/render_tree.py for rendering code.
class ToyTreeMark(Mark):
    """Tree Mark optionally including Node markers and tip labels.

    """
    def __init__(self, **kwargs):
        Mark.__init__(self, annotation=False)

        # relevant for Mark domain and extents
        self._coordinate_axes = ['x', 'y']
        """: names of the Cartesian axes."""
        self.ntable: np.ndarray = kwargs.get("ntable")
        """: coordinates of the Nodes."""
        self.etable: np.ndarray = kwargs.get("etable")
        """: 2D array of edge idx labels."""
        self.ttable: np.ndarray = kwargs.get("ttable")
        """: coordinates of tip Nodes (diff when tip_labels_align=True)"""

        # all validated style values
        for key, val in kwargs.items():
            setattr(self, key, val)

    @property
    def nnodes(self) -> int:
        """The number of nodes in the tree"""
        return self.ntable.shape[0]

    def domain(self, axis: str) -> Tuple[float, float]:
        """The Nodes define the domain of data (tip labels not included).

        The Cartesian axes will only place ticks across the data domain.
        """
        if self.layout[0] == "c":
            domain = toyplot.data.minimax(self.ntable[:, :])
            absdomain = max(abs(i) for i in domain)
            return -absdomain, absdomain
        index = self._coordinate_axes.index(axis)
        domain = toyplot.data.minimax(self.ntable[:, index])
        return domain

    def extents(self, axis: Union[str, Sequence[str]]) -> Tuple[Tuple[np.ndarray], Tuple[np.ndarray]]:
        """The tip labels and Node markers define the extents of the data.

        This is extra padding to ensure node markers, e.g., circle doesn't
        get cut off by the axes. But most importantly it is to create
        padding to fit the tip labels. To do so requires parsing info
        about the tip labels style dict.

        Note
        ----
        Extents are [min-x, max-x, min-y, max-y]; remember that larger
        y-values are down on the canvas.

        Returns
        -------
        coords: Tuple[np.ndarray, np.ndarray]
        extents: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        Each ndarray is len=nnodes.
        """
        # this array can be modified during rendering so grab a copy.
        # get node coords but set tip-align coords.
        coords = self.ntable.copy()
        coords[:self.ttable.shape[0], :] = self.ttable

        # select coords on 'x', 'y' or ['x', 'y'] axes.
        coords = (coords[:, 0], coords[:, 1])

        # get marker/label extents and update minmax for each feature
        extents = [np.zeros(self.nnodes)] * 4
        extents = set_node_label_extents(self, extents)
        extents = set_marker_extents(self, extents)
        extents = set_edge_extents(self, extents)
        extents = set_tip_label_extents(self, extents)

        # self._coordinate_axes[i] for in list(axis)
        return coords, extents


def set_node_label_extents(mark: Mark, extents: List[np.ndarray]) -> List[np.ndarray]:
    """Return extents of node label text string extents.
    """
    # return if no labels
    if mark.node_labels is None:
        return extents

    # else return the calculated text extents
    ext = toyplot.text.extents(
        text=mark.node_labels,
        angle=0,
        style=mark.node_labels_style
    )
    extents[0] = np.min([extents[0], ext[0]], axis=0)
    extents[1] = np.max([extents[1], ext[1]], axis=0)
    extents[2] = np.min([extents[2], ext[2]], axis=0)
    extents[3] = np.max([extents[3], ext[3]], axis=0)
    return extents


def set_marker_extents(mark: Mark, extents: List[np.ndarray]) -> List[np.ndarray]:
    """Return extents of node label text string extents.
    """
    # return if no nodes
    if not any(mark.node_sizes):
        return extents

    # markers are symmetrical NxN unless 'rNxM'
    yext = np.zeros(mark.nnodes)
    xext = np.zeros(mark.nnodes)

    iter_data = zip(mark.node_markers, mark.node_sizes)
    for idx, (marker, size) in enumerate(iter_data):
        if marker.startswith("r"):
            width, height = [float(i) for i in marker[1:].split("x")]
        else:
            width, height = 1, 1
        yext[idx] = height * size
        xext[idx] = width * size

    # extent is half the markers dimension in either direction
    xext = xext / 2. + mark.node_style["stroke-width"]
    yext = yext / 2. + mark.node_style["stroke-width"]

    # set extents
    extents[0] = np.min([extents[0], -xext], axis=0)
    extents[1] = np.max([extents[1], xext], axis=0)
    extents[2] = np.min([extents[2], -yext], axis=0)
    extents[3] = np.max([extents[3], yext], axis=0)
    return extents


def set_edge_extents(mark: Mark, extents: List[np.ndarray]) -> List[np.ndarray]:
    """Return extents of edge stroke widths (usually small)."""
    if mark.edge_widths is None:
        widths = np.repeat(mark.edge_style["stroke-width"] / 2., mark.nnodes)
    else:
        widths = mark.edge_widths / 2.

    extents[0] = np.min([extents[0], -widths], axis=0)
    extents[1] = np.max([extents[1], widths], axis=0)
    extents[2] = np.min([extents[2], -widths], axis=0)
    extents[3] = np.max([extents[3], widths], axis=0)
    return extents


def set_tip_label_extents(mark: Mark, extents: List[np.ndarray]) -> List[np.ndarray]:
    """Return extents for each tip label text string.

    TODO
    ----
    - check -180 on 'ud'
    - support 'c', 'unr'
    - censor vertical text extent
    - support 'shrink' space
    """
    # return if not tip labels
    if mark.tip_labels is None:
        return extents

    # add layout-based angles and styles.
    angles = mark.tip_labels_angles
    if mark.layout in ['u', 'l']:
        angles = mark.tip_labels_angles - 180

    # else return the calculated text extents
    ext = toyplot.text.extents(
        text=mark.tip_labels,
        angle=angles,
        style=mark.tip_labels_style
    )

    # only concerned with tip Nodes
    ntips = len(mark.tip_labels)

    # TEMPORARY HACK: until toyplot extents is fixed.
    # extend the tip label direction extra space.
    if mark.layout == "r":
        ext[1][:ntips] *= 2
        ext[1][:ntips] += mark.shrink
    elif mark.layout == "d":
        ext[3][:ntips] *= 2
        ext[3][:ntips] += mark.shrink
    elif mark.layout == "u":
        ext[2][:ntips] *= 2
        ext[2][:ntips] -= mark.shrink
    elif mark.layout == "l":
        ext[0][:ntips] *= 2
        ext[0][:ntips] -= mark.shrink

    # Not important b/c we use fit-range later.
    elif mark.layout[0] == "c":
        ext[0][:ntips] *= 1.5
        ext[1][:ntips] *= 1.5
        ext[2][:ntips] *= 1.5
        ext[3][:ntips] *= 1.5
        ext[0][:ntips] -= mark.shrink

    # elif mark.layout[:3] == "unr":
        # logger.debug("unrooted layout needs custom tip ext!")

    # only allow increasing extents
    extents[0][:ntips] = np.min([extents[0][:ntips], ext[0]], axis=0)
    extents[1][:ntips] = np.max([extents[1][:ntips], ext[1]], axis=0)
    extents[2][:ntips] = np.min([extents[2][:ntips], ext[2]], axis=0)
    extents[3][:ntips] = np.max([extents[3][:ntips], ext[3]], axis=0)
    return extents


if __name__ == "__main__":

    import toytree
    from toytree.style import validate_style
    from toytree.drawing.src.draw_toytree import (
        get_tree_style_base, get_layout, tree_style_to_css_dict)

    toytree.set_log_level("DEBUG")
    tree = toytree.rtree.rtree(5, seed=123)
    tree[2].name = "hello world"

    kwargs = tree.draw(
        debug=True,
        tip_labels=True,
        tip_labels_angles=0,
        tip_labels_style={'font-size': 15},
        node_sizes=16,
        node_mask=False,
    )

    style = get_tree_style_base(tree, tree_style=kwargs.pop("tree_style"))
    logger.warning(style.tip_labels_style.font_size)
    style = validate_style(tree, style, **kwargs)
    logger.warning(style.tip_labels_style.font_size)
    layout = get_layout(tree, style)
    logger.warning(style.tip_labels_style.font_size)

    ntable = layout.coords
    etable = tree.get_edges("idx")
    mark = ToyTreeMark(ntable=ntable, etable=etable, ttable=layout.tcoords, **tree_style_to_css_dict(style))
    logger.info(style.tip_labels_angles)
    print(f"domain x = {mark.domain('x')}")
    print(f"domain y = {mark.domain('y')}")
    print(mark.extents('x')[0])
    print(mark.extents('x')[1][0])
    print(mark.extents('x')[1][1])
    print(mark.extents('x')[1][2])
    print(mark.extents(['x', 'y'])[1][3])

    # mark = ToyTreeMark2(ntable=ntable, etable=etable, **tree_style_to_css_dict(style))
    # print(f"domain x = {mark.domain('x')}")
    # print(f"domain y = {mark.domain('y')}")
    # print(mark.extents('x')[0])
    # print(mark.extents('x')[1][0])
    # print(mark.extents('x')[1][1])
    # print(mark.extents('x')[1][2])
    # print(mark.extents(['x', 'y'])[1][3])



#!/usr/bin/env python

"""
Phylogenetic comparative methods functions made accessible at
:mod:`toytree.pcm` as a subpackage-level API.
"""

from typing import List, Optional
import pandas as pd
import numpy as np

from toytree.pcm.src.utils import get_vcv_matrix_from_tree
# from toytree.pcm.src.discrete_markov_model_sim import 
# from toytree.pcm.src.discrete_markov_model_sim import (
#     DiscreteMarkovModelSim,
#     DiscreteMarkovModelQMatrix,
# )


__all__ = [
    "get_vcv_from_tree",
    "get_qmatrix",
    "simulate_discrete_markov_data",
    "simulate_discrete_markov_data_on_tree",
]


def get_qmatrix_er(nstates: int, rate: float):
    pass

def get_qmatrix_sym(nstates: int, rates: List[float]):
    pass

def get_qmatrix_ard(nstates: int, rates: List[float]):
    pass


def get_instaneous_transition_rate_matrix(
    nstates: int,
    rates: List[float],
    ) -> pd.DataFrame:
    """Get an instantaneous transition rate matrix (Q matrix) from

    Parameters
    ----------
    nstates: int
        The number of states in the transition rate model.

    Returns
    -------
    pandas.DataFrame
        An instaneous transition rate matrix (Q matrix).

    Examples
    --------
    >>> # get a Q-matrix under a 3-state 'Equal-Rates' model
    >>> qmatrix = toytree.pcm.get_instaneous_transition_rate_matrix(3, 1e-3)
    """
    matrix = DiscreteMarkovModelQMatrix(nstates)
    qmat = matrix.get_rate_matrix(rates)
    return qmat


def simulate_discrete_markov_data(
    tree: 'ToyTree',
    qmatrix: np.ndarray,
    nsims: int = 1,
    ancestral_state: Optional[np.ndarray] = None,
    seed: Optional[int] = None,
    ) -> pd.DataFrame:
    """Return a DataFrame with simulated discrete trait values.

    The index (rows) are tip names and columns are the indices of
    nsims replicate simulations. Trait values for internal nodes 
    are not returned (see :func:`simulate_discrete_markov_data_on_tree`).
    Results can be treated like observed data, which is typically only
    available for tip nodes.

    Parameters
    ----------
    tree: toytree.ToyTree
        An input ToyTree on the edges of which the trait will be simulated.
    qmatrix: numpy.ndarray
        An instantaneous transition rate matrix between N states.
    nsims: int
        The number of replicate simulations to perform.
    seed: Optional[int]
        The numpy random number generator seed to use.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=10, treeheight=100, seed=123)
    >>> qmatrix = toytree.pcm.get_transition_rate_matrix(3, 1e-3)
    >>> data = toytree.pcm.simulate_discrete_markov_states(tree, qmatrix)

    See Also
    --------
    :func:`simulate_discrete_markov_data_on_tree`
        Return a list of ToyTree instances with simulated data 
        assigned to a feature named 'discrete' on every node.
    """
    model = DiscreteMarkovModelSim(tree, qmatrix, nsims, ancestral_state, seed)
    return pd.DataFrame(
        index=tree.get_tip_labels(), data=model._get_state_array().T)


def simulate_discrete_markov_data_on_tree(
    tree: 'ToyTree',
    qmatrix: np.ndarray,
    nsims: int = 1,
    ancestral_state: Optional[np.ndarray] = None,
    seed: int = None,
    ) -> List['ToyTree']:
    """Return a list of ToyTrees instance (or list of ToyTree instances) with
    simulated states assigned to every node of the tree.

    Parameters
    ----------
    tree:
        An input Toytree on the edges of which a trait will be simulated.
    qmatrix:
        An instantaneous transition rate matrix (Q matrix)
    nsims: int
        The number of replicates simulations to run.
    ancestral_state:
        ...
    seed: Optional[int]
        ...

    Returns
    -------
    Union[Tree, List[ToyTree]]
        One or more ToyTrees with a feature "discrete" assigned with
        a value to every node of the tree.

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=10, treeheight=100, seed=123)
    >>> qmatrix =
    >>> tree = toytree.pcm.simulate_discrete_markov_states_on_tree()
    """
    model = DiscreteMarkovModelSim(tree, qmatrix, nsims, ancestral_state, seed)
    return model._get_state_tree()



if __name__ == "__main__":

    import toytree
    TREE = toytree.rtree.unittree(10, treeheight=100, seed=123)

    # use a hand-made matrix
    ALPHA = 0.001
    BETA = 0.002
    QMATRIX = np.array([
        [-ALPHA, ALPHA],
        [BETA, -BETA],
    ])

    # or generate a symmetric matrix
    QMATRIX = DiscreteMarkovModelQMatrix(nstates=3).get_rate_matrix(5e-3)

    # show matrix and simulate
    print(QMATRIX)
    print(toytree.pcm.simulate_discrete_markov_data(TREE, QMATRIX, nsims=3))
    simtree = simulate_discrete_markov_data_on_tree(TREE, QMATRIX, nsims=1)
    print(simtree[0].get_node_data("discrete"))



#!/usr/bin/env python

"""Conversion of phylogeny to and from variance-covariance matrix.

"""

from typing import Union
import itertools
import numpy as np
import pandas as pd
import toytree
from toytree import ToyTree
from toytree.core.apis import add_subpackage_method, PhyloCompAPI


@add_subpackage_method(PhyloCompAPI)
def get_vcv_matrix_from_tree(
    tree: ToyTree,
    df: bool = False,
) -> Union[np.ndarray, pd.DataFrame]:
    """Return a variance-covariance matrix (DataFrame) from a ToyTree.

    The VCV represents the sum lengths of shared edges between
    pairs of samples as covariances (off-diagonals) and sum
    root-to-tip edge lengths of each sample as variances (diagonals)
    This matrix is often useful as it provides the expected variances
    and covariances of a continuous trait evolving on a tree under
    Brownian motion.

    Parameters
    ----------
    tree: toytree.ToyTree
        A tree on which to compute the VCV.
    df: bool
        True returns pandas DataFrame, else returns numpy ndarray.

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123, treeheight=3)
    >>> print(get_vcv_from_tree(tree))
    >>> #      r0     r1      r2      r3      r4
    >>> # r0  3.0     2.0     1.0     0.0     0.0
    >>> # r1  2.0     3.0     1.0     0.0     0.0
    >>> # r2  1.0     1.0     3.0     0.0     0.0
    >>> # r3  0.0     0.0     0.0     3.0     1.0
    >>> # r4  0.0     0.0     0.0     1.0     3.0
    """
    # get node distance matrix
    dmat = tree.distance.get_node_distance_matrix()

    # fill vcv array with shared dists (mrca to root)
    vcv = np.zeros((tree.ntips, tree.ntips))
    for tip1, tip2 in itertools.combinations(range(tree.ntips), 2):

        # get mrca node and its dist to root (co-variances)
        mrca = tree.get_mrca_node(tip1, tip2)
        vcv[tip1, tip2] = dmat[mrca.idx, tree.treenode.idx]
        vcv[tip2, tip1] = vcv[tip1, tip2]

    # fill diagonal with each tips dist
    for node in tree[:tree.ntips]:
        vcv[node._idx, node._idx] = node._dist

    # return as ndarray or dataframe
    if not df:
        return vcv

    # return as dataframe
    tlabels = tree.get_tip_labels()
    return pd.DataFrame(vcv, columns=tlabels, index=tlabels)


@add_subpackage_method(PhyloCompAPI)
def get_corr_matrix_from_tree(
    tree: ToyTree,
    df: bool = False,
) -> Union[np.ndarray, pd.DataFrame]:
    r"""Return a correlation matrix (DataFrame) from a ToyTree.

    The correlation matrix is computed from the variance-covariance
    matrix. The relationship between the VCV (C) and the correlation
    matrix (R) is:
    $$ R_{ij} = \frac{C_{ij}}{\sqrt{C_{ii} * C_{jj}}} $$

    Parameters
    ----------
    tree: toytree.ToyTree
        A tree on which to compute the correlation matrix.

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123, treeheight=3)
    >>> print(get_vcv_from_tree(tree))
    >>> #      r0     r1      r2      r3      r4
    >>> # r0  3.0     2.0     1.0     0.0     0.0
    >>> # r1  2.0     3.0     1.0     0.0     0.0
    >>> # r2  1.0     1.0     3.0     0.0     0.0
    >>> # r3  0.0     0.0     0.0     3.0     1.0
    >>> # r4  0.0     0.0     0.0     1.0     3.0
    """
    vcv = get_vcv_matrix_from_tree(tree)
    diag_std = np.sqrt(np.diag(vcv))
    outer = np.outer(diag_std, diag_std)
    corr = vcv / outer
    corr[vcv == 0] = 0
    if not df:
        return corr
    names = tree.get_tip_labels()
    return pd.DataFrame(corr, index=names, columns=names)


@add_subpackage_method(PhyloCompAPI)
def get_tree_from_vcv_matrix(vcv: Union[np.ndarray, pd.DataFrame]) -> ToyTree:
    """Return tree reconstructed from a variance-covariance matrix.

    This first converts the VCV, which represents unique and shared
    edge lengths, into a distance matrix by subtracting each element
    in the matrix by the max value in the matrix. Then neighbor
    joining is applied to the distance matrix. The returned tree
    is unrooted.

    Parameters
    ----------
    vcv: ArrayLike
        A variance-covariance matrix as a np.ndarray or pd.DataFrame.
    """
    max_value = np.array(vcv).max()
    return toytree.infer.infer_neighbor_joining_tree(max_value - vcv)


if __name__ == "__main__":

    tre = toytree.rtree.rtree(10, seed=123)
    vcv = get_vcv_matrix_from_tree(tre, df=True)
    print(tre)
    print(vcv)

    # ttt = get_tree_from_vcv(vcv)
    # print(ttt)



#!/usr/bin/env python

"""Maximum likelihood tree inference.

A didactic version with visualizations for learning ML. Not 
optimized for speed.
"""

from typing import Iterator, Callable, Generator, List

from abc import ABC, abstractmethod

from loguru import logger
from numpy.typing import ArrayLike
import numpy as np
import toyplot
import toytree

logger = logger.bind(name="toytree")

# not yet part of core toytree install
try:
    import sympy
except ImportError:
    # logger.warning("sympy is not installed.")
    pass

# pylint: disable=invalid-name

# BASE_ORDER = list("AGCT")
BASE_ORDER = list("TCAG")  # ziheng Yang
BASE_COLOR_SCHEME = ["blue", "green", "orange", "purple"]
AX0_x_width = 0.056  # 66
AX1_x_width = 3.6
AX1_y_width = 15
AX1_y_width_m = 1.3


class SubstitutionModel(ABC):
    """AbstractBaseClass for molecular substitution models.

    SubstitutionModel class objects are used to generate the rate 
    matrix and transition probability matrix, as well as the 
    (log)likelihood expression and function using sympy.
    """
    def __init__(self):
        self._PI = 0.25, 0.25, 0.25, 0.25
        """: default equilibrium frequencies of nucleotides for simple models"""
        self._q_matrix = sympy.Matrix()
        """: rate matrix assigned to the Model."""
        self._t = sympy.Symbol("t")
        """: the symbolic variable to express time span."""
        self._p_matrices = []
        """: probability matrices in from the intermediate (for class) to the simplified (for calculation) forms."""
        self._pairwise_likelihood_formula = None
        """: the likelihood under multinomial probability."""
        self._pairwise_loglikelihood_formula = None

    @property
    def q_matrix(self) -> sympy.Matrix:
        """Return the rate matrix (Q) given the model parameters."""
        return self._q_matrix

    @q_matrix.setter
    def q_matrix(self, value: sympy.Matrix) -> None:
        """Set a value to the q_matrix property attribute."""
        self._q_matrix = value

    def p_matrix(self, intermediate_id: int = -1) -> sympy.Matrix:
        """Return the transition probability matrix (P) given model parameters.
        
        Parameters
        ----------
        intermediate_id: int
            The values at intermediate steps can be returned and viewed
            for didactic purposes, if specified here.
        """
        if not self._p_matrices:
            self._update_p_matrix()
        return self._p_matrices[intermediate_id]

    def get_pairwise_likelihood_formula(self, logarithm: bool = True) -> sympy.Mul:
        """Return the likelihood formula for the substitution model.

        This function uses sympy ...
        """
        if not self._pairwise_loglikelihood_formula:
            self._update_pairwise_likelihood_formula()
        if logarithm:
            return self._pairwise_loglikelihood_formula
        return self._pairwise_likelihood_formula

    @abstractmethod
    def _update_p_matrix(self):
        """Child classes have a proper function here."""        

    @abstractmethod
    def _update_pairwise_likelihood_formula(self):
        """Child classes have a proper function here."""

    @abstractmethod
    def get_p_matrix_function(self):
        """Return a function to get P matrix given Q and d."""

    @abstractmethod
    def get_pairwise_loglike_function(self, *args):
        """Return a function to return likelihood given a distance."""


class JC69(SubstitutionModel):
    """Jukes-Cantor substitution Model."""
    def __init__(self):
        # init parent class
        super().__init__()

        # add new attributes of this SubstitutionModel child class
        self._mu = mu = sympy.Symbol("mu")
        """: symbolic variable for instantaneous rate of change to any base"""
        self._d = sympy.Symbol("d")
        """: symbolic variable for phylogenetic distance, with d=3*mu*t"""
        self._n = sympy.Symbol("n")
        """: symbolic variable for observed total number of sites"""
        self._x = sympy.Symbol("x")
        """: symbolic variable for observed number of differences"""
        self._q_matrix = sympy.Matrix([[-3 * mu, mu, mu, mu],
                                       [mu, -3 * mu, mu, mu],
                                       [mu, mu, -3 * mu, mu],
                                       [mu, mu, mu, -3 * mu]])
        """: the rate matrix (Q)"""

    def _update_p_matrix(self):
        """Append P matrix intermediates for didactic purposes."""
        self._p_matrices.append((self._q_matrix * self._t).exp())
        self._p_matrices.append(self._p_matrices[0].subs(self._mu*self._t, self._d/3))

    def _update_pairwise_likelihood_formula(self):
        """Sets function to attribute `._pairwise_(log)likelihood_formula.

        Considering multinomial distribution for all 16 site patterns,
        the probability is p_0/4 for any constant site pattern
        the probability is p_1/4 for any different site pattern
        given that those site patterns apply to all four bases
        """        
        p_matrix = self.p_matrix()
        self._pairwise_likelihood_formula = \
            (p_matrix[0, 1] * self._PI[1]) ** self._x * \
            (p_matrix[0, 0] * self._PI[0]) ** (self._n - self._x)
        self._pairwise_loglikelihood_formula = \
            self._x * sympy.log(p_matrix[0, 1] * self._PI[1]) + \
            (self._n - self._x) * sympy.log(p_matrix[0, 0] * self._PI[0])

    def get_pairwise_loglike_function(
        self,
        change_matrix: ArrayLike = None,
        x_obs: int = None,
        n_obs: int = None,
        ) -> sympy.core.function:
        """Return a function to return likelihood given a distance.

        Generates a function that takes phylogenetic distance (d) as 
        its parameter and returns the (log?)-likelihood value. If x
        and n are input, the change_matrix will be skipped.

        Parameters
        ----------
        change_matrix: np.array
            observed matrix of changes in the form of:
            >>>                 species 2
            >>>                 --------------
            >>>                 A   G   C   T
            >>> species 1 | A  n1  n2  n3  n4
            >>>           | G  n5  n6  n7  n8
            >>>           | C  n9  n10 n11 n12
            >>>           | T  n13 n14 n15 n16
            >>>
            >>> e.g. example from Z.H. Yang (2014)
            >>> np.array([[179,  23,   1,   0],
            >>>           [ 30, 219,   2,   0],
            >>>           [  2,   1, 291,  10],
            >>>           [  0,   0,  21, 169]])
        x_obs : int
            observed number of differences
        n_obs : int
            observed total number of sites

        Returns
        -------
        sympy.core.function
            a function that takes phylogenetic distance (d) as its 
            parameter and returns a log-likelihood float value.
        """
        if (x_obs and n_obs) or change_matrix is not None:
            if not (x_obs and n_obs):
                n_obs = change_matrix.sum()
                x_obs = n_obs - change_matrix.diagonal()
            loglike_formula = self.get_pairwise_likelihood_formula(logarithm=True)
            return sympy.lambdify(
                [self._d],
                loglike_formula.evalf(subs={self._n: n_obs, self._x: x_obs})
            )
        raise TypeError("Please input change_matrix or x_obs and n_obs!")

    def get_p_matrix_function(self) -> Callable:
        """Return a function to calculate transition probability matrix
        for a given distance (d), used to calculate likelihood.

        Note
        ----
        This is much faster than simply using 
        `return sympy.lambdify([self._d], self.p_matrix())`

        Returns
        -------
        sympy.core.function
            a function that takes phylogenetic distance (d: float) and
            kappa_ls (optional) as its parameter(s), and outputs the 
            probability matrix (ArrayLike)
        """
        p_matrix = self.p_matrix()
        p0_f = sympy.lambdify([self._d], p_matrix[0, 0], modules="numpy")
        p1_f = sympy.lambdify([self._d], p_matrix[0, 1], modules="numpy")

        def _p_matrix_quick_func(_d, _kappa):
            p0, p1, = p0_f(_d, _kappa), p1_f(_d, _kappa)
            return np.array([[p0, p1, p1, p1],
                             [p1, p0, p1, p1],
                             [p1, p1, p0, p1],
                             [p1, p1, p1, p0]])
        return _p_matrix_quick_func


class K80(SubstitutionModel):
    """Kimura 1980 Substitution Model.

    This model has separate rates for transitions (alpha) and 
    tranversions (beta), specified by the parameters kappa=alpha/beta.
    """
    def __init__(self):
        super().__init__()
        self._alpha = alpha = sympy.Symbol("alpha")
        """: the symbolic variable to store the rate of transitions, i.e. A <-> G, C <-> T"""
        self._beta = beta = sympy.Symbol("beta")
        """: the symbolic variable to store the rate of transversions, i.e. A, G <-> C, T"""
        self._kappa = sympy.Symbol("kappa_ls")
        """: the symbolic variable to store transition/transversion rate ratio, i.e. alpha/beta"""
        self._d = sympy.Symbol("d")
        """: the symbolic variable to store phylogenetic distance, with d=(alpha+2*beta)*t"""
        self._n = sympy.Symbol("n")
        """: the symbolic variable to store observed total number of sites"""
        self._s = sympy.Symbol("S")
        """: the symbolic variable to store observed transitions"""
        self._v = sympy.Symbol("V")
        """: the symbolic variable to store observed transversions"""
        self._q_matrix = sympy.Matrix([[-(alpha+2*beta), alpha, beta, beta],
                                       [alpha, -(alpha+2*beta), beta, beta],
                                       [beta, beta, -(alpha+2*beta), alpha],
                                       [beta, beta, alpha, -(alpha+2*beta)]])
        """: the rate matrix"""
        self._mu = alpha+2*beta
        """: the average substitution rate"""

    def _update_p_matrix(self):
        """Append P matrix intermediates for didactic purposes."""        
        # original probability matrix
        self._p_matrices.append((self._q_matrix * self._t).exp())

        ###################################################
        # --- use d and kappa_ls to replace alpha and beta ---
        # 1. get the expression of alpha and beta using phylogenetic distance, timespan, kappa_ls
        solution = sympy.solve(
            [self._d - (self._alpha+2*self._beta) * self._t,
             self._kappa - self._alpha / self._beta],
            [self._alpha,
             self._beta])

        # 2. replace alpha and beta
        self._p_matrices.append(
            sympy.simplify(
                self._p_matrices[0].
                subs(self._alpha, solution[self._alpha]).
                subs(self._beta, solution[self._beta])
            )
        )


    def _update_pairwise_likelihood_formula(self):
        """...

        # considering multinomial distribution for all 16 site patterns, the probabilities are
        # p_0/4 for any constant site pattern
        # p_1/4 for any transitional site pattern
        # p_2/4 for any transversional site pattern, respectively
        """
        p_matrix = self.p_matrix()
        self._pairwise_likelihood_formula = \
            (p_matrix[0, 0] * self._PI[0]) ** (self._n - self._s - self._v) * \
            (p_matrix[0, 1] * self._PI[1]) ** self._s * \
            (p_matrix[0, 2] * self._PI[2]) ** self._v
        self._pairwise_loglikelihood_formula = \
            (self._n - self._s - self._v) * sympy.log(p_matrix[0, 0] * self._PI[0]) + \
            self._s * sympy.log(p_matrix[0, 1] * self._PI[0]) + \
            self._v * sympy.log(p_matrix[0, 2] * self._PI[0])

    def get_pairwise_loglike_function(self,
                                      change_matrix: ArrayLike = None,
                                      s_obs: int = None,
                                      v_obs: int = None,
                                      n_obs: int = None) -> sympy.core.function:
        """
        This function generates a function that takes phylogenetic distance (d) as its parameter,
        and outputs the log-likelihood value.
        If x and n are input, change_matrix will be skipped.

        Parameters
        ----------
        change_matrix: np.array
            observed matrix of changes in the form of
                            species 2
                            --------------
                            A   G   C   T
            species 1 | A  n1  n2  n3  n4
                      | G  n5  n6  n7  n8
                      | C  n9  n10 n11 n12
                      | T  n13 n14 n15 n16
            e.g. example from Z.H. Yang (2014)
            np.array([[179,  23,   1,   0],
                      [ 30, 219,   2,   0],
                      [  2,   1, 291,  10],
                      [  0,   0,  21, 169]])

        s_obs : int
            observed transitional sites
        v_obs : int
            observed transversional sites
        n_obs : int
            observed total number of sites
        Returns
        -------
        sympy.core.function
            a function that takes phylogenetic distance (d) as its parameter and outputs the log-likelihood value
        """
        if (s_obs and v_obs and n_obs) or change_matrix is not None:
            if not (s_obs and v_obs and n_obs):
                n_obs = change_matrix.sum()
                s_obs = change_matrix[0, 1] + change_matrix[1, 0] + change_matrix[2, 3] + change_matrix[3, 2]
                v_obs = n_obs - change_matrix.diagonal().sum() - s_obs
            loglike_formula = self.get_pairwise_likelihood_formula()
            return sympy.lambdify([self._d, self._kappa],
                                  loglike_formula.evalf(subs={self._n: n_obs, self._s: s_obs, self._v: v_obs}))
        else:
            raise TypeError("Please input change_matrix, or s_obs, v_obs and n_obs!")

    def get_p_matrix_function(self, kappa: float = 2.):
        """
        The function used in maximum likelihood tree inference

        Parameters
        ----------
        kappa: float
            transitional sites over transversional sites.
            The resulting function will be ca. 15% faster when kappa_ls is fixed

        Returns
        -------
        sympy.core.function
            a function that takes phylogenetic distance (d: float) and kappa_ls (optional) as its parameter(s),
            and outputs the probability matrix (ArrayLike)
        """
        # three times faster than simply using `return sympy.lambdify([self._d, self._kappa], self.p_matrix())`
        # even slightly faster (ca. 15%) than a straightforward function based on the formula
        p_matrix = self.p_matrix()
        p0_f = sympy.lambdify([self._d], p_matrix[0, 0].subs(self._kappa, kappa), modules="numpy")
        p1_f = sympy.lambdify([self._d], p_matrix[0, 1].subs(self._kappa, kappa), modules="numpy")
        p2_f = sympy.lambdify([self._d], p_matrix[0, 2].subs(self._kappa, kappa), modules="numpy")
        if kappa is not None:
            def _p_matrix_quick_func(_d):
                p0, p1, p2 = p0_f(_d), p1_f(_d), p2_f(_d)
                return np.array([[p0, p1, p2, p2],
                                 [p1, p0, p2, p2],
                                 [p2, p2, p0, p1],
                                 [p2, p2, p1, p0]])
        else:
            def _p_matrix_quick_func(_d, _kappa):
                p0, p1, p2 = p0_f(_d, _kappa), p1_f(_d, _kappa), p2_f(_d, _kappa)
                return np.array([[p0, p1, p2, p2],
                                 [p1, p0, p2, p2],
                                 [p2, p2, p0, p1],
                                 [p2, p2, p1, p0]])
        return _p_matrix_quick_func


class TN93(SubstitutionModel):
    """Tamura and Nei (1993) Substitution Model.

    Combines different equilibrium frequency distributions with unequal
    transition (alpha) and transversion (beta) rates, as well as 
    different rates for transitions between pyrimadines (alpha1) and 
    purines (alpha2).
    """
    def __init__(self):
        super().__init__()
        self._PI = pi_1, pi_2, pi_3, pi_4 = sympy.symbols(["pi_" + base_type for base_type in BASE_ORDER])
        """: using symbolic variables to rewrite the equilibrium frequencies of nucleotides"""
        self._alpha_1 = alpha_1 = sympy.Symbol("alpha_1")
        """: the symbolic variable to store the rate of transitions, A<->G"""
        self._alpha_2 = alpha_2 = sympy.Symbol("alpha_2")
        """: the symbolic variable to store the rate of transitions, T<->C"""
        self._beta = beta = sympy.Symbol("beta")
        """: the symbolic variable to store the rate of transversions, i.e. A, G <-> C, T"""
        self._kappa_1 = sympy.Symbol("kappa_1")
        """: the symbolic variable to store alpha1/beta"""
        self._kappa_2 = sympy.Symbol("kappa_2")
        """: the symbolic variable to store alpha2/beta"""
        self._d = sympy.Symbol("d")
        """: the symbolic variable to store phylogenetic distance, with d=(alpha+2*beta)*t"""
        self._change_mtx = sympy.Matrix([["O_" + from_b + to_b for from_b in BASE_ORDER] for to_b in BASE_ORDER])
        """: the symbolic matrix to store observed change of matrix"""
        self._q_matrix = sympy.Matrix([[-(pi_2*alpha_1+pi_3*beta+pi_4*beta), pi_1*alpha_1, pi_1*beta, pi_1*beta],
                                       [pi_2*alpha_1, -(pi_1*alpha_1+pi_3*beta+pi_4*beta), pi_2*beta, pi_2*beta],
                                       [pi_3*beta, pi_3*beta, -(pi_1*beta+pi_2*beta+pi_4*alpha_2), pi_3*alpha_2],
                                       [pi_4*beta, pi_4*beta, pi_4*alpha_2, -(pi_1*beta+pi_2*beta+pi_3*alpha_2)]])
        """: the rate matrix"""
        # not the following form
        # self._q_matrix = sympy.Matrix([[-alpha_1*pi_2-beta*pi_3-beta*pi_4, alpha_1*pi_2, beta*pi_3, beta*pi_4],
        #                                [alpha_1*pi_1, -alpha_1*pi_1 - beta*pi_3 - beta*pi_4, beta*pi_3, beta*pi_4],
        #                                [beta*pi_1, beta*pi_2, -alpha_2*pi_4 - beta*pi_2 - beta*pi_1, alpha_2*pi_4],
        #                                [beta*pi_1, beta*pi_2, alpha_2*pi_3, -alpha_2*pi_3 - beta*pi_2 - beta*pi_1]])
        # self._express = None
        # """: alpha/beta -> f(d,t,kappa_ls)"""
        self._p_matrix_w = None
        self._log_p_matrix_w = None
        """: tmp variables"""
        self._mu = -(self._q_matrix.diagonal() * sympy.Matrix(self._PI))[0]
        """: substitution rate"""

    def _update_express(self):
        self._express = sympy.solve(
            [self._d - self._mu * self._t,
             self._kappa_1 - self._alpha_1 / self._beta,
             self._kappa_2 - self._alpha_2 / self._beta],
            [self._alpha_1,
             self._alpha_2,
             self._beta])

    # def _update_p_matrix(self):
    #     # original probability matrix
    #     self._p_matrices.append((self._q_matrix * self._t).exp())
    #     self._reparametrize_p_matrix()
    #
    # def _reparametrize_p_matrix(self):
    #     # --- use d and kappa_ls to replace alpha and beta ---
    #     # 1. get the expression of alpha_1, alpha_2, beta using phylogenetic distance, timespan, kappa_1, kappa_2
    #     if not self._express:
    #         self._update_express()
    #     # # 2. replace alpha_1, alpha_2, beta
    #     self._p_matrices.append(sympy.simplify(self._p_matrices[-1].
    #                                            subs(self._alpha_1, self._express[self._alpha_1]).
    #                                            subs(self._alpha_2, self._express[self._alpha_2]).
    #                                            subs(self._beta, self._express[self._beta])))

    def _update_p_matrix(self):
        """
        Instead of using self._p_matrices.append((self._q_matrix * self._t).exp())
        we have the p matrix stored to speed up
        """
        pi1, pi2, pi3, pi4 = self._PI
        d = self._d
        b = self._beta
        a1 = self._alpha_1
        a2 = self._alpha_2
        k1 = self._kappa_1
        k2 = self._kappa_2
        t = self._t
        exp = sympy.exp
        # to shorten the expression, do semi-manual combinations in the formula
        pi12 = pi1+pi2   # pyridines or purines
        pi34 = pi3+pi4   # purines or pyridines
        c1 = 2*(k1*pi2*pi1+k2*pi3*pi4+pi12*pi34)
        self._p_matrices.append(
            sympy.Matrix(
                [[(pi3*pi1+pi4*pi1)*exp(-b*t)/pi12+pi2*exp(-a1*pi12*t-b*pi34*t)/pi12+pi1,
                  (pi3*pi1+pi4*pi1)*exp(-b*t)/pi12-pi1*exp(-a1*pi12*t-b*pi34*t)/pi12+pi1,
                  pi1-pi1*exp(-b*t),
                  pi1-pi1*exp(-b*t)],
                 [pi2-pi2*exp(-a1*pi12*t-b*pi34*t)/pi12+(pi3*pi2+pi2*pi4)*exp(-b*t)/pi12,
                  pi2+(pi3*pi2+pi2*pi4)*exp(-b*t)/pi12+pi1*exp(-a1*pi12*t-b*pi34*t)/pi12,
                  pi2-pi2*exp(-b*t),
                  pi2-pi2*exp(-b*t)],
                 [-pi3*exp(-b*t)+pi3,
                  -pi3*exp(-b*t)+pi3,
                  pi3+pi3*pi12*exp(-b*t)/pi34-exp(-a2*pi34*t-b*pi12*t)/(-pi3/pi4-1),
                  pi3-pi3*exp(-a2*pi34*t-b*pi12*t)/pi34+pi3*pi12*exp(-b*t)/pi34],
                 [-pi4*exp(-b*t)+pi4,
                  -pi4*exp(-b*t)+pi4,
                  pi4+pi4*pi12*exp(-b*t)/pi34+exp(-a2*pi34*t-b*pi12*t)/(-pi3/pi4-1),
                  pi3*exp(-a2*pi34*t-b*pi12*t)/pi34+pi4+pi4*pi12*exp(-b*t)/pi34]]
            )
        )

        self._p_matrices.append(
            sympy.Matrix(
                [[pi1+(pi2*exp(d*(-k1*pi12-pi34+1)/c1)+pi1*pi34)*exp(-d/c1)/pi12,
                  pi1+pi1*(pi34-exp(d*(-k1*pi12-pi34+1)/c1))*exp(-d/c1)/pi12,
                  pi1-pi1*exp(-d/c1),
                  pi1-pi1*exp(-d/c1)],
                 [pi2+pi2*(pi34-exp(d*(-k1*pi12-pi34+1)/c1))*exp(-d/c1)/pi12,
                  pi2+(pi2*pi34+pi1*exp(d*(-k1*pi12-pi34+1)/c1))*exp(-d/c1)/pi12,
                  pi2-pi2*exp(-d/c1),
                  pi2-pi2*exp(-d/c1)],
                 [pi3-pi3*exp(-d/c1),
                  pi3-pi3*exp(-d/c1),
                  pi3+(pi3*pi12+pi4*exp(d*(-k2*pi34-pi12+1)/c1))*exp(-d/c1)/pi34,
                  pi3+pi3*(pi12-exp(d*(-k2*pi34-pi12+1)/c1))*exp(-d/c1)/pi34],
                 [pi4-pi4*exp(-d/c1),
                  pi4-pi4*exp(-d/c1),
                  pi4+pi4*(pi12-exp(d*(-k2*pi34-pi12+1)/c1))*exp(-d/c1)/pi34,
                  pi4+(pi3*exp(d*(-k2*pi34-pi12+1)/c1)+pi4*pi12)*exp(-d/c1)/pi34]]
            )
        )

    def _update_pairwise_likelihood_formula(self):
        """considering multinomial distribution for all 16 site patterns"""
        # np.array and sympy.Matrix has different multiplication scheme
        self._p_matrix_w = np.array(self.p_matrix()) * np.array(self._PI)
        self._log_p_matrix_w = np.array(sympy.simplify(sympy.Matrix(self._p_matrix_w).applyfunc(sympy.log)))
        self._pairwise_likelihood_formula = self._p_matrix_w ** np.array(self._change_mtx)
        self._pairwise_loglikelihood_formula = np.array(self._change_mtx) * self._log_p_matrix_w

    def get_pairwise_loglike_function(self, change_matrix: ArrayLike) -> sympy.core.function:
        """
        This function generates a function that takes phylogenetic distance (d) as its parameter,
        and outputs the log-likelihood value.
        If x and n are input, change_matrix will be skipped.

        Parameters
        ----------
        change_matrix: np.array
            observed matrix of changes in the form of
                            species 2
                            --------------
                            A   G   C   T
            species 1 | A  n1  n2  n3  n4
                      | G  n5  n6  n7  n8
                      | C  n9  n10 n11 n12
                      | T  n13 n14 n15 n16
            e.g. example from Z.H. Yang (2014)
            np.array([[179,  23,   1,   0],
                      [ 30, 219,   2,   0],
                      [  2,   1, 291,  10],
                      [  0,   0,  21, 169]])

        Returns
        -------
        sympy.core.function
            a function that takes phylogenetic distance (d) as its parameter and outputs the log-likelihood value
        """
        if self._p_matrix_w is None:
            self._update_pairwise_likelihood_formula()
        # here we did not use the self._pairwise_likelihood_formula (directly RES_LIKELIHOOD JC69 and K80)
        # because the generated function will be slow if two many separate parameters were used
        # instead, we use the self._log_p_matrix_w and the array of change_matrix to generate the formula
        like_formula = (self._log_p_matrix_w * change_matrix).sum()
        return sympy.lambdify([self._d, self._kappa_1, self._kappa_2] + self._PI, like_formula)

    # TODO using Horner's rule may help speed up, i.e. combining terms in common
    def get_p_matrix_function(self,
                              kappa_ls: tuple[float, float] = None,
                              pi_ls: tuple[float, float, float, float] = None):
        """
        The function used in maximum likelihood tree inference.
        The resulting function will be faster when more parameters are fixed.

        Parameters
        ----------
        kappa_ls: tuple[float, float]
            To which kappa_1 (alpha_1/beta) and kappa_2 (alpha_2/beta) are fixed.
        pi_ls: tuple[float, float, float, float]
            To which equilibrium frequencies of nucleotides are fixed.

        Returns
        -------
        sympy.core.function
            a function that takes phylogenetic distance (d: float) and kappa (optional) as its parameter(s),
            and outputs the probability matrix (ArrayLike)
        """
        # much faster than simply using `return sympy.lambdify([self._d, self._kappa], self.p_matrix())`
        p_matrix = self.p_matrix()
        if kappa_ls is not None and pi_ls is not None:
            # fastest
            p_matrix = p_matrix.evalf(
                subs={self._kappa_1: kappa_ls[0],self._kappa_2: kappa_ls[1]} |
                     {self._PI[go_b]: pi_val for go_b, pi_val in enumerate(pi_ls)})
            return sympy.lambdify([self._d], p_matrix)
        elif kappa_ls:
            p_matrix = p_matrix.evalf(subs={self._kappa_1: kappa_ls[0], self._kappa_2: kappa_ls[1]})
            return sympy.lambdify([self._d] + self._PI, p_matrix)
        elif pi_ls:
            p_matrix = p_matrix.evalf(subs={self._PI[go_b]: pi_val for go_b, pi_val in enumerate(pi_ls)})
            return sympy.lambdify([self._d, self._kappa_1, self._kappa_2], p_matrix)
        else:
            return sympy.lambdify([self._d, self._kappa_1, self._kappa_2] + self._PI, p_matrix)


def node_conditional_probability(
    node: toytree.Node,
    p_matrix_func: sympy.core.function,
) -> ArrayLike:
    """Return the conditional probability of a Node.

    This is a recursive function that can be called on the root Node
    of a tree to get the conditional likelihood of the whole tree. It
    calls itself on all children of each internal Node until it reaches
    the tips of the tree.
    """
    if node.is_leaf():
        return node.conditional_prob

    # internal Node computes conditional from child values
    child_cd_prob_generator = (
        node_conditional_probability(child_node, p_matrix_func)
        for child_node in node.children)
    prob_matrix_generator = (
        p_matrix_func(child_node.dist)
        for child_node in node.children)
    node.conditional_prob = combine_descendent_conditional_probability(
        child_conditional_prob_gen=child_cd_prob_generator,
        child_prob_matrix_gen=prob_matrix_generator)
    return node.conditional_prob


def combine_descendent_conditional_probability(
    child_conditional_prob_gen: Iterator[ArrayLike],
    child_prob_matrix_gen: Iterator[ArrayLike],
) -> ArrayLike:
    r"""Return conditional likelihood of a Node.

    Calculates the conditional probability of current node given its
    descendant conditional probability list and corresponding 
    transition probability matrix.

    Under non-logarithm mode, for node $i$ with descendent nodes {$j$} 
    and for each base type x, it calculates:

    $L_i(x_i)$ = $\product_j \sum_{x_j} p_{{x_i}{x_j}}(t_j)L_j(x_j)$

    the product of the 'transferred conditional probability' from all 
    descendent nodes. The 'transferred conditional probability' 
    ($\sum_{x_j} p_{{x_i}{x_j}}(t_j)L_j(x_j)$) here describe the sum 
    over conditional probabilities that base type x may accumulate from
    all base types from a descendent node.

    Parameters
    ----------
    child_conditional_prob_gen : Iterator[ArrayLike in the shape of (4,)]
        Input iterator that each time generates the conditional probability of a child.
        Each conditional probability array should be of `shape=(4,)`, e.g. `np.array([0.05, 0.001, 0.04, 0.002])`
    child_prob_matrix_gen: Iterator[ArrayLike in the shape of (4,4)]
        Input iterator that each time generates the transition probability matrix of a child.
        Each transition probability matrix should be of `shape=(4,4)`,
        e.g. `np.array([[0.8644351 , 0.06591888, 0.03482301, 0.03482301],
                        [0.06591888, 0.8644351 , 0.03482301, 0.03482301],
                        [0.03482301, 0.03482301, 0.8644351 , 0.06591888],
                        [0.03482301, 0.03482301, 0.06591888, 0.8644351 ]])` # K80 TEST_MODEL with d=0.15, k=2.
        This can be generated using the function of `get_p_matrix_function` under each substitution TEST_MODEL objects.

    Returns
    -------
    ArrayLike[float]
        Output the conditional probability of the combined use np.array in the shape of (4,).
    """
    accumulated_prob = (next(child_conditional_prob_gen) * next(child_prob_matrix_gen)).sum(axis=1)
    for child_prob in child_conditional_prob_gen:
        this_p_matrix = next(child_prob_matrix_gen)
        accumulated_prob *= (child_prob * this_p_matrix).sum(axis=1)
    return accumulated_prob


def get_tree_likelihood(
    tree: toytree.ToyTree,
    data: dict[str: str],
    p_matrix_func: sympy.core.function,
    pi_list: ArrayLike,
) -> float:
    """Return the log-likelihood of observing data given a tree.

    """
    # set observed data as features of tip Nodes
    for nidx in range(tree.ntips):
        node = tree[nidx]
        assert node.name in data, f"data must include all tips in the tree. Missing={node.name}"
        state = data[node.name]
        node.conditional_prob = np.array(BASE_ORDER) == state

    # infer internal node conditional probs from tip states
    root_prob = node_conditional_probability(tree.treenode, p_matrix_func)
    return (root_prob * pi_list).sum()


######################################################
#
# VISUALIZATION FUNCTIONS
#
######################################################


def combine_descendent_conditional_probability_with_plot(
    child_conditional_prob_list: List[ArrayLike],
    child_prob_matrix_list: List[ArrayLike],
    ax1,
) -> ArrayLike:
    """Return visualization of combine_descendent_conditional_probability.

    Parameters
    ----------
    child_conditional_prob_list : Iterator[ArrayLike in the shape of (4,)]
        Input iterator that each time generates the conditional probability
        of a child. Each conditional probability array should be of 
        `shape=(4,)`, e.g. `np.array([0.05, 0.001, 0.04, 0.002])`
    child_prob_matrix_list: Iterator[ArrayLike in the shape of (4,4)]
        Input iterator that each time generates the transition probability 
        matrix of a child. Each transition probability matrix should be of 
        `shape=(4,4)`, e.g.:
        >>> np.array([[0.8644351 , 0.06591888, 0.03482301, 0.03482301],
        >>>           [0.06591888, 0.8644351 , 0.03482301, 0.03482301],
        >>>           [0.03482301, 0.03482301, 0.8644351 , 0.06591888],
        >>>           [0.03482301, 0.03482301, 0.06591888, 0.8644351 ]])`
        >>>           # K80 TEST_MODEL with d=0.15, k=2.
        This can be generated using the function of `get_p_matrix_function`
        under each substitution TEST_MODEL objects.
    ax1: toyplot.Axis
        ...

    Returns
    -------
    ArrayLike[float]
        Output the conditional probability of the combined use np.array
        in the shape of (4,).
    """
    # plot input values
    for go_base in range(4):
        # conditional probs
        x_coords_c = [40 - (1.5 - go_base) * AX1_x_width for foo in child_conditional_prob_list]
        y_coords_c = [20 - (0.5 - go_c) * AX1_y_width for go_c, foo in enumerate(child_conditional_prob_list)]
        ax1.scatterplot(
            x_coords_c,
            y_coords_c,
            size=15,
            marker=[
                toyplot.marker.create(shape='r3.9x1',
                                      label=f'{child_conditional_prob[go_base]:.7f}',
                                      size=11,
                                      mstyle={"fill": BASE_COLOR_SCHEME[go_base], "stroke": "lightgrey",
                                              "fill-opacity": 0.2})
                for child_conditional_prob in child_conditional_prob_list
            ]
        )
        # p matrix
        x_coords_m = [40 - (1.5 - _to_base) * AX1_x_width
                      for foo in child_conditional_prob_list
                      for _to_base in range(4)]
        y_coords_m = [20 - (0.5 - go_c) * AX1_y_width - (3 + go_base) * AX1_y_width_m
                      for go_c, foo in enumerate(child_conditional_prob_list)
                      for _to_base in range(4)]
        ax1.scatterplot(
            x_coords_m,
            y_coords_m,
            size=15,
            marker=[
                toyplot.marker.create(shape='r3.9x1',
                                      label=f'{p_m[go_base][_to_base]:.7f}',
                                      size=11,
                                      mstyle={"fill": BASE_COLOR_SCHEME[_to_base], "stroke": "lightgrey",
                                              "fill-opacity": 0.2})
                for p_m in child_prob_matrix_list
                for _to_base in range(4)
            ]
        )
        # heads of p matrix
        labels = [[from_to + _base for _base in BASE_ORDER] for from_to in ("to ", "from ")]
        ax1.scatterplot(
            [40 + from_to * AX1_x_width * 2.5 - (1.5 - go_base) * AX1_x_width * (1 - from_to)
             for foo in child_conditional_prob_list
             for from_to in range(2)],
            [20 - (0.5 - go_c) * AX1_y_width - (3 + go_base) * AX1_y_width_m * from_to - 2 * (
                    1 - from_to) * AX1_y_width_m
             for go_c, foo in enumerate(child_conditional_prob_list)
             for from_to in range(2)],
            size=15,
            marker=[
                toyplot.marker.create(shape='r3.9x1',
                                      label=f'{labels[from_to][go_base]}',
                                      size=11,
                                      mstyle={"fill": BASE_COLOR_SCHEME[go_base], "stroke": "white",
                                              "fill-opacity": 0.05})
                for foo in child_conditional_prob_list
                for from_to in range(2)
            ]
        )

    # calculate the intermediate matrices
    intermediate_matrices = []
    for go_child, child_conditional_probability in enumerate(child_conditional_prob_list):
        intermediate_matrices.append(child_conditional_probability * child_prob_matrix_list[go_child])
    # plot the intermediate matrices
    for go_base in range(4):
        x_coords_m = [23.5 - (1.5 - _to_base) * AX1_x_width
                      for foo in child_conditional_prob_list
                      for _to_base in range(4)]
        y_coords_m = [20 - (0.4 - go_c) * AX1_y_width - (3 + go_base) * AX1_y_width_m
                      for go_c, foo in enumerate(child_conditional_prob_list)
                      for _to_base in range(4)]
        ax1.scatterplot(
            x_coords_m,
            y_coords_m,
            size=15,
            marker=[
                toyplot.marker.create(shape='r3.9x1',
                                      label=f'{intermediate_m[go_base][_to_base]:.7f}',
                                      size=11,
                                      mstyle={"fill": BASE_COLOR_SCHEME[_to_base], "stroke": "lightgrey",
                                              "fill-opacity": 0.2})
                for intermediate_m in intermediate_matrices
                for _to_base in range(4)
            ]
        )

    # calculate the sums
    summed_matrices = []
    for intermediate_matrix in intermediate_matrices:
        summed_matrices.append(intermediate_matrix.sum(axis=1))
    # plot the sums
    for go_base in range(4):
        # sum
        ax1.scatterplot(
            [23.5 - 2.7 * AX1_x_width for foo in child_conditional_prob_list],
            [20 - (0.4 - go_c) * AX1_y_width - (3 + go_base) * AX1_y_width_m for go_c, foo in
             enumerate(child_conditional_prob_list)],
            size=15,
            marker=[
                toyplot.marker.create(shape='r3.9x1',
                                      label=f'{sum_matrix[go_base]:.7f}',
                                      size=11,
                                      mstyle={"fill": BASE_COLOR_SCHEME[go_base], "stroke": "lightgrey",
                                              "fill-opacity": 0.2})
                for sum_matrix in summed_matrices
            ]
        )

    # calculate the product of the sums
    accumulated_prob = summed_matrices[0]
    for summing_matrix in summed_matrices[1:]:
        accumulated_prob *= summing_matrix
    # plot the final result
    for go_base in range(4):
        x_coords_m = [4 - (1.5 - go_base) * AX1_x_width]
        y_coords_m = [20 - 3 * AX1_y_width_m]
        ax1.scatterplot(x_coords_m, y_coords_m, size=15, marker=[
            toyplot.marker.create(shape='r3.9x1',
                                  label=f'{accumulated_prob[go_base]:.7f}',
                                  size=11,
                                  mstyle={"fill": BASE_COLOR_SCHEME[go_base], "stroke": "lightgrey",
                                          "fill-opacity": 0.2})
        ]
    )
    return accumulated_prob


def node_conditional_probability_with_plot(
    node, p_matrix_func: sympy.core.function, tree: toytree.ToyTree,
) -> Generator:
    """More description...

    Parameters
    ----------
    node: toytree.Node
        ...
    p_matrix_func: Callable
        ...
    tree: toytree.ToyTree
        ...
    """
    if node.is_leaf():
        yield node.conditional_prob, ""
    else:
        # create the generators, recursively
        child_cd_prob_list = []
        child_prob_matrix_list = []
        for child_node in node.children:
            child_generator = node_conditional_probability_with_plot(child_node, p_matrix_func, tree)
            conditional_prob, prob_canvas = next(child_generator)
            yield conditional_prob, prob_canvas
            prob_matrix = p_matrix_func(child_node.dist)
            for conditional_prob, prob_canvas in child_generator:
                yield conditional_prob, prob_canvas
            child_cd_prob_list.append(conditional_prob)
            child_prob_matrix_list.append(prob_matrix)
        # create the canvas
        node.conditional_prob_canvas = toyplot.Canvas(width=1200, height=400)
        ax0 = node.conditional_prob_ax0 = node.conditional_prob_canvas.cartesian(
            bounds=(50, 450, 50, 350), padding=40)
        ax1 = node.conditional_prob_ax1 = node.conditional_prob_canvas.cartesian(
            bounds=(550, 1150, 50, 350), padding=50, xmin=0, xmax=50, ymin=0, ymax=30)
        ax0.show = False
        ax1.show = False
        # plot tree
        tree.style.edge_colors = ['black'] * tree.nnodes
        for child_node in node.children:
            tree.style.edge_colors[child_node.idx] = "red"
        tree.draw(node_labels=True, node_sizes=12, width=30, height=400, axes=ax0)
        node.conditional_prob = combine_descendent_conditional_probability_with_plot(
            child_conditional_prob_list=child_cd_prob_list,
            child_prob_matrix_list=child_prob_matrix_list,
            ax1=ax1)
        # plot generated prob
        for go_base in range(4):
            # plot probs along the tree
            candidate_probs = tree.get_node_data("conditional_prob")
            accessible_points = [bool(len(j)) for j in candidate_probs]
            accessible_probs = [j for go_p, j in enumerate(candidate_probs) if accessible_points[go_p]]
            coords = tree.get_node_coordinates()[accessible_points]
            ax0.scatterplot(
                coords.x - (1.5 - go_base) * AX0_x_width,
                coords.y - 0.2,
                size=15,
                marker=[
                    toyplot.marker.create(shape='r3.9x1',
                                          label=f"{i[go_base]:.7f}",
                                          size=11,
                                          mstyle={"fill": BASE_COLOR_SCHEME[go_base], "stroke": "lightgrey",
                                                  "fill-opacity": 0.2})
                    for i in accessible_probs
                ]
            )
        yield node.conditional_prob, node.conditional_prob_canvas


def get_tree_likelihood_plot_gen(
    tree: toytree.ToyTree,
    tip_states: dict[str: str],
    p_matrix_func: sympy.core.function,
    pi_list: ArrayLike,
    ):
    """...
    """
    for sp_name, sp_state in tip_states.items():
        sp_node = tree.get_nodes(sp_name)[0]  # do we have better function to do this in toytree?
        sp_node.conditional_prob = (np.array(BASE_ORDER) == sp_state).astype(float)
    node_cp_plot_gen = node_conditional_probability_with_plot(tree.treenode, p_matrix_func=p_matrix_func, tree=tree)
    root_prob, current_canvas = next(node_cp_plot_gen)
    if current_canvas:
        yield "", root_prob, current_canvas
    for root_prob, current_canvas in node_cp_plot_gen:
        if current_canvas:
            yield "", root_prob, current_canvas
    yield (root_prob * pi_list).sum(), root_prob, tree.treenode.conditional_prob_canvas



if __name__ == "__main__":

    TEST_ARRAY = np.array([[179,  23,   1,   0],
                           [ 30, 219,   2,   0],
                           [  2,   1, 291,  10],
                           [  0,   0,  21, 169]])
    OBSERVED_DATA = {"sp-A": "T", "sp-B": "C", "sp-C": "A", "sp-D": "C", "sp-E": "C"}
    TREE = toytree.tree("(((sp-A:0.2,sp-B:0.2):0.1,sp-C:0.2):0.1,(sp-D:0.2,sp-E:0.2):0.1);")
    # TEST_MODEL = K80()
    TEST_MODEL = JC69()
    TEST_MODEL.p_matrix()
    print(TEST_MODEL._p_matrices)
    print(TEST_MODEL.q_matrix)

    # LIK = get_tree_likelihood(tree=TREE,
    #                           data=OBSERVED_DATA,
    #                           p_matrix_func=TEST_MODEL.get_p_matrix_function(kappa=2.),
    #                           pi_list=TEST_MODEL._PI)
    # print(LIK)
    # TREE.treenode.draw_ascii()



#!/usr/bin/env python

"""Testing simple alternative validation

Problem
-------
[x] function calls from Tree
[x] default/object style from Tree
[x] Annotate module only has nnodes, etc.
[ ] communicate node_colors(single) -> node_style['fill']
[ ] all substyles as dicts w/ keys converted to css style: (fill-opacity, -toyplot-anchor-shift)
"""

from typing import Union, Sequence, Mapping, Any, Dict, TypeVar
import numpy as np
from loguru import logger
import toyplot
from toytree.color import ToyColor
from toytree.utils import ToytreeError
from toytree.style.src.validate_utils import check_arr
from toytree.style.src.style_base import TipLabelStyle, TreeStyle
from toytree.style.src.map_colors import get_color_mapped_values

logger = logger.bind(name="toytree")
ToyTree = TypeVar("ToyTree")
Color = TypeVar("Color", str, tuple, np.ndarray)

__all__ = [
    "validate_tip_labels",
    "validate_tip_labels_colors",
    "validate_tip_labels_angles",
    "validate_tip_labels_style",
]


def validate_tip_labels(
    tree: ToyTree,
    style: TreeStyle,
    **kwargs,
) -> Union[None, np.ndarray]:
    """Sets tip_labels to np.ndarray[str] or None.

    Also applies floating point string formatting on tip_labels. If
    set to None then no labels are added during rendering.

    tip_labels: Union[None, bool, str, Sequence[str]],

    Args
    ----
    None: None
    False: None
    True: 'idx' int labels
    str: extract feature
    Series[Any: custom collection of the proper size.
    """
    # get user value or style base value
    tip_labels = kwargs.get("tip_labels")
    if tip_labels is None:
        tip_labels = getattr(style, "tip_labels")

    # Don't use 'is in' to support pd.Series
    if tip_labels is False:
        return None
    if tip_labels is None:
        return None
    if tip_labels is True:
        return tree.get_tip_labels()

    # str: extract feature as labels as an np.array
    if isinstance(tip_labels, str):
        tip_labels = tree.get_tip_data(tip_labels).values

    # or, user entered a list, tuple, Series, array, etc.
    # which is checked and handled below.

    # float format numeric values, and convert to strings
    tip_labels = [str(i) for i in tip_labels]
    return check_arr(tip_labels, "tip_labels", tree.ntips, str)


def validate_tip_labels_colors(
    tree: ToyTree,
    style: TreeStyle,
    **kwargs,
) -> Union[np.ndarray, None]:
    """Return ndarray of tip_colors or None.

    If only one color was entered then return None and the
    tip_labels_style dict will instead set the color as 'fill'.

    colors: Union[None, Color, Sequence[Color], tuple[str, str]],

    Args
    ----
    None: None. Node color will use node_style.fill
    str: set all as ToyColor(str)
    ndarray dtype: set all as ToyColor(ndarray)
    tuple
    colormap: not supported. Use (feature, colormap)
    """
    # get user value or style base value
    colors = kwargs.get("tip_labels_colors")
    if colors is None:
        if isinstance(style, TreeStyle):
            colors = getattr(style, "tip_labels_colors")

    # if None then Nodes will be colored by style.fill during render
    if colors is None:
        return None, None

    # if None then Nodes will be colored by style.fill during render
    if colors is None:
        return None, None

    # special (feature, cmap), but not single color as tuple (0, 0, 1, 1).
    if isinstance(colors, tuple) and isinstance(colors[0], str):
        feat, *args = colors
        values = tree.get_tip_data(feat).values
        return get_color_mapped_values(values, *args), None

    # get user entered value as ToyColor or List[ndarray]
    # e.g., str, tuple, or ndarray[r,g,b,a] -> ToyColor
    # e.g., List[str], pd.Series[ToyColor], ndarray[void] -> List[ToyColor]
    colors = ToyColor.color_expander(colors)

    # if single color return None. The node_style validator will recognize
    # None and check orig node_color arg to override node_style.fill.
    if isinstance(colors, ToyColor):
        return None, colors

    # convert the List[ToyColor] to an ndarray of dtype np.void
    colors = check_arr(
        values=colors,
        label="tip_labels_colors",
        size=tree.ntips,
        ctype=np.void,
    )

    # serialize is optionally used...
    # if serialize:
    #     return list(tip_labels_colors)
    return colors, None


# def validate_tip_labels_font_sizes(
#     tree: ToyTree,
#     sizes: Union[float, Sequence[float]],
# ) -> np.ndarray:
#     """Sets node_sizes to ndarray[float]."""
#     if sizes is None:
#         sizes = 0
#     if isinstance(sizes, (int, float)):
#         sizes = np.repeat(sizes, tree.ntips)
#     sizes = check_arr(
#         sizes, "tip_labels_font_sizes", tree.ntips, (int, float, np.integer))
#     return sizes


def validate_tip_labels_angles(
    tree: ToyTree,
    style: TreeStyle,
    **kwargs,
) -> np.ndarray:
    """Return array with tip label angles given the layout, or user setting.
    """
    # get user value or style base value
    angles = kwargs.get("tip_labels_angles")
    if angles is None:
        angles = getattr(style, "tip_labels_angles")

    # get user value or style base value
    layout = kwargs.get("layout")
    if not layout:
        layout = getattr(style, "layout")

    # calculate propor angles from the layout
    if angles is None:
        if layout in ("u", "d"):  ################## also support kwargs layout
            angles = np.repeat(-90, tree.ntips)
        else:
            angles = np.zeros(tree.ntips)
    elif isinstance(angles, (float, int, np.integer)):
        angles = np.repeat(angles, tree.ntips)

    angles = check_arr(angles, "tip_labels_angles", tree.ntips, (int, float, np.integer))
    return angles


def validate_tip_labels_style(
    tree: ToyTree,
    style: TreeStyle = None,
    **kwargs,
) -> Dict[str, Any]:
    """Ensure tip labels are in px sizes and fill color as ToyColor.

    Parameters
    ----------
    base: TreeStyle or None
        A base tree style to modify. If None then new NodeLabelStyle is used.
    style: Mapping
        A dict of style changes.
    copy: bool
        If False the base is copied and modified, else modified directly.
    """
    # annotation funcs use builtin base
    if style is None:
        style = TipLabelStyle()

    # update default style with user style args
    for key, val in kwargs.items():
        style[key] = val

    # ensure font-size is in px units
    size = toyplot.units.convert(value=style.font_size, target="px", default="px")
    style.font_size = f"{size:.12g}px"

    # convert fill to a ToyColor
    style.fill = ToyColor(style.fill)
    return style


if __name__ == "__main__":

    pass



#!/usr/bin/env python

"""..."""

from toytree.style.src.style_base import TreeStyle, SubStyle
from toytree.style.src.style_types import get_base_tree_style_by_name
from toytree.style.src.map_colors import (
    get_color_mapped_feature, get_color_mapped_values)
from toytree.style.src.map_values import (
    get_range_mapped_feature, get_range_mapped_values)
from toytree.style.src.validate_style import validate_style
from toytree.style.src.validate_utils import (
    tree_style_to_css_dict, substyle_dict_to_css_dict, check_arr)

__all__ = [
    "TreeStyle",
    "SubStyle",
    "get_base_tree_style_by_name",
    "get_color_mapped_feature",
    "get_color_mapped_values",
    "get_range_mapped_feature",
    "get_range_mapped_values",
    "validate_style",
    "tree_style_to_css_dict",
    "substyle_dict_to_css_dict",
    "check_arr",
]

if __name__ == "__main__":

    import toytree
    tre = toytree.rtree.unittree(5)
    sty = tre.style

    sty.node_colors = ["red"] * tre.nnodes
    sty.tip_labels_colors = toytree.color.COLORS1[0]
    sty.node_markers = "o"

    # validate_style(tre, sty)
    print(sty)



#!/usr/bin/env python

"""Scripts to run comparison against chronos in R.
"""

import subprocess
import numpy as np
import toyplot

RSTRING = """
library(ape)

# load the Python variables
btree <- read.tree(text="{tree}")
min_ages <- c{min_ages}
max_ages <- c{max_ages}
tips <- list{tips}
lamb <- {lamb}
model <- '{model}'

# run the R code 
nodes <- c()
for (tipset in tips) {{
    mrca <- getMRCA(btree, tipset)
    nodes <- append(nodes, mrca)
}}

calib <- data.frame(node=nodes, age.min=min_ages, age.max=max_ages)

ctree <- chronos(
    btree, 
    lambda=lamb,
    model=model,
    calibration=calib,
)
write.tree(ctree)
"""


def convert_to_rvec(tup):
    return "c('" + "','".join(tup) + "')"

def _run_chronos_in_r(tree, lamb=1, model="relaxed", calibrations=None):
    """
    For testing and validating pen.lik. methods
    """
    calibrations = calibrations if calibrations is not None else {}
    if not calibrations:
        calibrations = {tree.treenode.idx: (1, 1)}
    cmin_age = [str(calibrations[i][0]) for i in sorted(calibrations)]
    cmax_age = [str(calibrations[i][1]) for i in sorted(calibrations)]

    data = {
        "tree": tree.write(),
        "min_ages": "(" + ",".join(cmin_age) + ")",
        "max_ages": "(" + ",".join(cmax_age) + ")",
        "tips": "(" + ",".join(
            convert_to_rvec(tree.get_tip_labels(x))
            for x in sorted(calibrations)
        ) + ")" if calibrations else "()",
        "lamb": lamb,
        "model": model,
    }
    rst = RSTRING.format(**data)
    with open("/tmp/chronos.R", 'w') as out:
        out.write(rst)
    out = subprocess.run(
        ["Rscript", "/tmp/chronos.R"], 
        check=True,
        stdout=subprocess.PIPE,
    )
    out = out.stdout.decode()
    ctre = out.strip().split("[1] ")[-1].strip('"')
    return ctre, out



class TreeSampler:
    """
    Class for applying uncorrelated gamma rate transformations to edges
    to create Generative data for testing the Chronos functions.
    """
    def __init__(self, tree, neff=5e5, gentime=1, gamma=3):
        self.tree = tree
        self.neff = neff
        self.gentime = gentime
        self.gamma = gamma
        self.nnodes = self.tree.nnodes


    def plot_gamma_distributed_rates(self, nsamples=10000, bins=30):
        """
        Draws the stat distribution for verification.
        """
        a = (1 / self.gamma) ** 2
        b = 1
        gamma_rates = np.random.gamma(shape=1 / (a*b**2), scale=a*b**2, size=nsamples)

        # generate a distribution of G or Ne values
        NEVAR = gamma_rates * self.neff
        canvas = toyplot.Canvas(width=600, height=250)
        ax0 = canvas.cartesian(grid=(1, 2, 0), xlabel="gamma distributed Ne variation")
        m0 = ax0.bars(np.histogram(NEVAR, bins=bins))

        GVAR = gamma_rates * self.gentime
        ax1 = canvas.cartesian(grid=(1, 2, 1), xlabel="gamma distributed gentime variation")
        m1 = ax1.bars(np.histogram(GVAR, bins=bins))
        return canvas, (ax0, ax1), (m0, m1)
    
    
    def get_tree(self, N=False, G=False, transform=False, seed=None):
        """
        N (bool):
            Sample gamma distributed variation in Ne values across nodes.
        G (bool):
            Sample gamma distributed variation in G values across nodes.
        transform (int):
            Return edge lengths in requested units by selecting integer value:
            (0) time, (1) coalunits, (2) generations.    
        """
        # get a copy of the sptree
        tree = self.tree.copy()
        
        # use random seed
        if seed:
            np.random.seed(seed)
        
        # get gamma rate multipliers
        a = (1 / self.gamma) ** 2
        b = 1
        
        # sample N values
        if N:
            gamma_rates = np.random.gamma(
                shape=1 / (a*b**2), 
                scale=a*b**2, 
                size=self.nnodes,
            )

            # sample Ne values from gamma dist
            nevals = gamma_rates * self.neff
        
            # apply randomly to nodes of the tree
            tree = tree.set_node_values(
                feature="Ne",
                mapping={i: nevals[i] for i in range(tree.nnodes)}
            )
        else:
            tree = tree.set_node_values("Ne", default=self.neff)
        
        if G:
            gamma_rates = np.random.gamma(
                shape=1 / (a*b**2), 
                scale=a*b**2, 
                size=self.nnodes,
            )

            # sample Ne values from gamma dist
            gvals = gamma_rates * self.gentime
        
            # apply randomly to nodes of the tree
            tree = tree.set_node_values(
                feature="g",
                mapping={i: gvals[i] for i in range(tree.nnodes)}
            )
        else:
            tree = tree.set_node_values("g", default=self.gentime)      
        
        # optionally convert edges to coal units
        if transform == 1:
            tree = tree.set_node_values(
                feature="dist",
                mapping={i: j.dist / (j.Ne * 2 * j.g) for i,j in enumerate(tree)}
            )
            
        elif transform == 2:
            tree = tree.set_node_values(
                feature="dist",
                mapping={i: j.dist / j.g for i,j in enumerate(tree)}
            )            
        return tree


if __name__ == "__main__":

    import toytree
    TREE = toytree.rtree.unittree(10, treeheight=1e6)
    TSA = TreeSampler(TREE, neff=5e5, gentime=1, gamma=3)
    TRE = TSA.get_tree()
    print(TRE)
